"""Tool for finding and characterizing lobes in antenna radiation patterns."""

from __future__ import annotations

from math import isclose, isnan
from typing import Optional

import numpy as np
import pydantic.v1 as pd
from pandas import DataFrame

from tidy3d.components.base import Tidy3dBaseModel, cached_property, skip_if_fields_missing
from tidy3d.components.types import ArrayFloat1D, ArrayLike, Ax
from tidy3d.constants import fp_eps
from tidy3d.exceptions import ValidationError
from tidy3d.log import log

from .viz import plot_params_lobe_FNBW, plot_params_lobe_peak, plot_params_lobe_width

# The minimum plateau size for peak finding, which is set to 0 to ensure that all peaks are found.
# A value must be provided to retrieve additional information from `find_peaks`.
MIN_PLATEAU_SIZE = 0
DEFAULT_MIN_LOBE_REL_HEIGHT = 1e-3
DEFAULT_NULL_THRESHOLD = 1e-3


class LobeMeasurer(Tidy3dBaseModel):
    """
    Tool for detecting and analyzing lobes in antenna radiation patterns,
    along with their characteristics such as direction and beamwidth.

    Example
    -------
    >>> theta = np.linspace(0, 2 * np.pi, 301, endpoint=False)
    >>> Urad = np.cos(theta) ** 2 * np.cos(3 * theta) ** 2
    >>> lobe_measurer = LobeMeasurer(
    ...     angle=theta,
    ...     radiation_pattern=Urad) # doctest: +SKIP
    >>> lobe_measures = lobe_measurer.lobe_measures # doctest: +SKIP
    """

    angle: ArrayFloat1D = pd.Field(
        ...,
        title="Angle",
        description="A 1-dimensional array of angles in radians. The angles should be "
        "in the range [0, 2π] and should be sorted in ascending order.",
    )

    radiation_pattern: ArrayFloat1D = pd.Field(
        ...,
        title="Radiation Pattern",
        description="A 1-dimensional array of real values representing the radiation pattern "
        "of the antenna measured on a linear scale.",
    )

    apply_cyclic_extension: bool = pd.Field(
        True,
        title="Apply Cyclic Extension",
        description="To enable accurate peak finding near boundaries of the ``angle`` array, "
        "we need to extend the signal using its periodicity. If lobes near the boundaries are not "
        "of interest, this can be set to ``False``.",
    )

    width_measure: float = pd.Field(
        0.5,
        gt=0.0,
        le=1.0,
        title="Beamwidth Measure",
        description="Relative magnitude of the lobes at which the beamwidth is measured. "
        "Default value of ``0.5`` corresponds with the half-power beamwidth.",
    )

    min_lobe_height: float = pd.Field(
        DEFAULT_MIN_LOBE_REL_HEIGHT,
        gt=0.0,
        le=1.0,
        title="Minimum Lobe Height",
        description="Only lobes in the radiation pattern with heights above this value are found. "
        "Lobe heights are measured relative to the maximum value in ``radiation_pattern``.",
    )

    null_threshold: float = pd.Field(
        DEFAULT_NULL_THRESHOLD,
        gt=0.0,
        le=1.0,
        title="Null Threshold",
        description="The threshold for detecting nulls, "
        "which is relative to the maximum value in the ``radiation_pattern``.",
    )

    @pd.validator("angle", always=True)
    def _sorted_angle(cls, val):
        """Ensure the angle array is sorted."""
        if not np.all(np.diff(val) >= 0):
            raise ValidationError("The angle array must be sorted in ascending order.")
        return val

    @pd.validator("radiation_pattern", always=True)
    def _nonnegative_radiation_pattern(cls, val):
        """Ensure the radiation pattern is nonnegative."""
        if not np.all(val >= 0):
            raise ValidationError("Radiation pattern must be nonnegative.")
        return val

    @pd.validator("apply_cyclic_extension", always=True)
    @skip_if_fields_missing(["angle"])
    def _cyclic_extension_valid(cls, val, values):
        if val:
            angle = values.get("angle")
            if np.any(angle < 0) or np.any(angle > 2 * np.pi):
                raise ValidationError(
                    "When using cyclic extension, the angle array must be in the range [0, 2π]."
                )
        return val

    @cached_property
    def lobe_measures(self) -> DataFrame:
        """
        The lobe measures as a pandas ``pandas.DataFrame`` with the following columns:

        - ``direction``: The angular position of the lobe peak.
        - ``magnitude``: The height of the lobe peak.
        - ``beamwidth``: The width of the lobe at the specified beamwidth measure.
        - ``beamwidth magnitude``: The magnitude at which the beamwidth is measured.
        - ``beamwidth bounds``: The bounds of the lobe at the the beamwidth measure.
        - ``FNBW``: The first null beam width (FNBW) of the lobe.
        - ``FNBW bounds``: The locations of the nulls on either side of the lobe.

        Returns
        -------
        DataFrame
            A DataFrame containing all lobe measures, where rows indicate the lobe index.
        """
        from scipy.signal import find_peaks

        if self.apply_cyclic_extension:
            angle, signal = self.cyclic_extension(self.angle, self.radiation_pattern)
        else:
            angle, signal = self.angle, self.radiation_pattern

        peaks, peak_props = find_peaks(
            signal,
            plateau_size=MIN_PLATEAU_SIZE,
            height=self.min_lobe_height,
        )
        if len(peaks) == 0:
            return DataFrame()
        # Find the locations of nulls in the signal
        max_mag = np.max(signal)
        null_height = self.null_threshold * max_mag
        _, null_peak_props = find_peaks(
            -signal,
            plateau_size=MIN_PLATEAU_SIZE,
            height=-null_height,
        )

        # Extract and post process the peak properties from Scipy
        peak_locations = self._extract_peak_locations(angle, peak_props)
        peak_heights = self._extract_peak_heights(peak_props)
        widths, beamwidth_height, beamwidth_min, beamwidth_max = self._calc_peak_widths(
            angle, signal, peaks
        )
        null_locations = self._extract_peak_locations(angle, null_peak_props)
        FNBWs, FNBW_min, FNBW_max = self._first_null_beam_widths(peak_locations, null_locations)

        width_bounds = list(zip(beamwidth_min, beamwidth_max))
        FNBW_bounds = list(zip(FNBW_min, FNBW_max))
        data = {
            "direction": peak_locations,
            "magnitude": peak_heights,
            "beamwidth": widths,
            "beamwidth magnitude": beamwidth_height,
            "beamwidth bounds": width_bounds,
            "FNBW": FNBWs,
            "FNBW bounds": FNBW_bounds,
        }

        dataframe = DataFrame(data)
        if self.apply_cyclic_extension:
            dataframe = self._filter_out_of_bounds_lobes(dataframe)
        return dataframe

    @staticmethod
    def cyclic_extension(angle: ArrayLike, signal: ArrayLike) -> tuple[ArrayLike, ArrayLike]:
        """
        This helper function extends the given `signal` array by leveraging its periodic nature.
        It ensures that the peak finding algorithm in Scipy can reliably detect peaks near the
        minimum and maximum values of the ``angle`` array. The returned arrays are extended to
        the range [-π, 3π).

        Parameters
        ----------
        angle : ArrayLike
            The array of angles, expected to be in the range [0, 2π].
        signal : ArrayLike
            The array of signal values corresponding to the angle array.

        Returns
        -------
        tuple[ArrayLike, ArrayLike]
            A tuple containing the extended `angle` and `signal` arrays.
        """
        # Ensure the last sample is not duplicated if it is at 2π
        if isclose(angle[-1], 2 * np.pi, rel_tol=fp_eps):
            angle = angle[:-1]
            signal = signal[:-1]

        first_locs = np.array(np.logical_and(angle >= 0, angle < np.pi)).nonzero()
        second_locs = np.array(np.logical_and(angle >= np.pi, angle < 2 * np.pi)).nonzero()
        angle = np.concatenate(
            [angle[second_locs] - 2 * np.pi, angle, angle[first_locs] + 2 * np.pi]
        )
        signal = np.concatenate([signal[second_locs], signal, signal[first_locs]])
        return angle, signal

    @staticmethod
    def _extract_peak_locations(angle: ArrayLike, peak_props: dict) -> ArrayLike:
        """Get the peak positions in terms of the angular coordinates."""
        left_locs = peak_props["left_edges"]
        right_locs = peak_props["right_edges"]
        left_x = angle[left_locs]
        right_x = angle[right_locs]
        peaks = (left_x + right_x) / 2
        return peaks

    @staticmethod
    def _extract_peak_heights(peak_props: dict) -> ArrayLike:
        """Get the peak heights."""
        return peak_props["peak_heights"]

    def _calc_peak_widths(
        self, angle: ArrayLike, signal: ArrayLike, peaks: ArrayLike
    ) -> tuple[ArrayLike, ArrayLike, ArrayLike, ArrayLike]:
        """Get the peak widths in terms of the angular coordinates."""
        from scipy.signal import peak_widths

        rel_height = 1.0 - self.width_measure
        last_element = len(signal) - 1
        left_ips = np.zeros_like(peaks)
        right_ips = last_element * np.ones_like(peaks)
        dummy_prominence_data = (signal[peaks], left_ips, right_ips)
        widths, width_heights, left_ips, right_ips = peak_widths(
            signal, peaks, rel_height, dummy_prominence_data
        )
        left_x = np.interp(left_ips, np.arange(len(angle)), angle)
        right_x = np.interp(right_ips, np.arange(len(angle)), angle)
        widths = right_x - left_x
        # Check that widths were properly found and are not simply the bounds
        improper_width_mask = np.logical_or(left_ips == 0, right_ips == last_element)
        widths[improper_width_mask] = np.nan
        width_heights[improper_width_mask] = np.nan
        left_x[improper_width_mask] = np.nan
        right_x[improper_width_mask] = np.nan
        return widths, width_heights, left_x, right_x

    @staticmethod
    def _first_null_beam_widths(
        peaks: ArrayLike, nulls: ArrayLike
    ) -> tuple[ArrayLike, ArrayLike, ArrayLike]:
        """Calculate the first null beam width for each lobe. If a beam null is not found,
        the value for the width is set to ``nan``."""
        insertion_indices = np.searchsorted(nulls, peaks)

        left_nulls = np.full_like(peaks, np.nan)
        right_nulls = np.full_like(peaks, np.nan)
        left_mask = (insertion_indices > 0) & (insertion_indices < len(nulls))
        right_mask = (insertion_indices >= 0) & (insertion_indices < len(nulls))
        left_nulls[left_mask] = nulls[insertion_indices[left_mask] - 1]
        right_nulls[right_mask] = nulls[insertion_indices[right_mask]]

        peak_FNBW = right_nulls - left_nulls
        return peak_FNBW, left_nulls, right_nulls

    @staticmethod
    def _filter_out_of_bounds_lobes(dataframe: DataFrame) -> DataFrame:
        """Filter out lobes that are out of bounds of the original signal."""
        filtered_dataframe = dataframe[
            dataframe["direction"].between(0, 2 * np.pi, inclusive="left")
        ]
        return filtered_dataframe.reset_index(drop=True)

    @property
    def main_lobe(self) -> DataFrame:
        """Properties of the main lobe."""
        if self.lobe_measures.empty:
            return DataFrame()
        main_lobe_idx = self.lobe_measures["magnitude"].idxmax()
        return self.lobe_measures.iloc[main_lobe_idx]

    @property
    def side_lobes(self) -> DataFrame:
        """Properties of all side lobes."""
        if self.lobe_measures.empty:
            return DataFrame()
        main_lobe_idx = self.lobe_measures["magnitude"].idxmax()
        side_lobes = self.lobe_measures.drop(main_lobe_idx)
        return side_lobes

    @property
    def sidelobe_level(self) -> Optional[float]:
        """The sidelobe level returned on a linear scale."""
        if self.side_lobes.empty:
            return None
        main_lobe_level = self.main_lobe["magnitude"]
        side_lobes = self.side_lobes
        max_side_lobe_idx = side_lobes["magnitude"].idxmax()
        side_lobe_level = side_lobes.at[max_side_lobe_idx, "magnitude"]
        return side_lobe_level / main_lobe_level

    def plot(
        self,
        lobe_index: int,
        ax: Ax,
        include_beamwidth: bool = True,
        include_FNWB: bool = True,
    ) -> Ax:
        """Annotate an existing ``matplotlib`` axis with lobe locations and widths.

        Parameters
        ----------
        lobe_index : int
            Index of the lobe to plot from the :attr:`lobe_measures`.
        ax : Ax
            ``matplotlib`` axis on which to plot the lobe measure.
        include_beamwidth : bool, optional
            If True, plot the beamwidth markers and fill the beamwidth region.
        include_FNWB : bool, optional
            If True, plot the First Null Beam Width markers and fill the FNBW region.

        Returns
        -------
        Ax
            The matplotlib axis with the plotted lobe measures.
        """

        lobe = self.lobe_measures.iloc[lobe_index]
        direction = lobe["direction"]
        ax.axvline(direction, **plot_params_lobe_peak.to_kwargs())

        # Additional annotations of the plot
        if include_beamwidth and not isnan(lobe["beamwidth magnitude"]):
            width_bounds = lobe["beamwidth bounds"]
            ax.axvline(width_bounds[0], **plot_params_lobe_width.to_kwargs())
            ax.axvline(width_bounds[1], **plot_params_lobe_width.to_kwargs())

        if include_FNWB and not isnan(lobe["FNBW"]):
            FNBW_bounds = lobe["FNBW bounds"]
            ax.axvline(FNBW_bounds[0], **plot_params_lobe_FNBW.to_kwargs())
            ax.axvline(FNBW_bounds[1], **plot_params_lobe_FNBW.to_kwargs())

        return ax

    @pd.root_validator(pre=False)
    def _warn_rf_license(cls, values):
        log.warning(
            "ℹ️ ⚠️ RF simulations are subject to new license requirements in the future. You have instantiated at least one RF-specific component.",
            log_once=True,
        )
        return values



"""Defines tetrahedral grid datasets."""

from __future__ import annotations

from typing import Union

import numpy as np
import pydantic.v1 as pd
from xarray import DataArray as XrDataArray

from tidy3d.components.base import cached_property
from tidy3d.components.data.data_array import (
    CellDataArray,
    IndexedDataArray,
    PointDataArray,
)
from tidy3d.components.types import ArrayLike, Axis, Bound, Coordinate
from tidy3d.exceptions import DataError
from tidy3d.packaging import requires_vtk, vtk

from .base import UnstructuredGridDataset
from .triangular import TriangularGridDataset


class TetrahedralGridDataset(UnstructuredGridDataset):
    """Dataset for storing tetrahedral grid data. Data values are associated with the nodes of
    the grid.

    Note
    ----
    To use full functionality of unstructured datasets one must install ``vtk`` package (``pip
    install tidy3d[vtk]`` or ``pip install vtk``). Otherwise the functionality of unstructured
    datasets is limited to creation, writing to/loading from a file, and arithmetic manipulations.

    Example
    -------
    >>> tet_grid_points = PointDataArray(
    ...     [[0.0, 0.0, 0.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
    ...     coords=dict(index=np.arange(4), axis=np.arange(3)),
    ... )
    >>>
    >>> tet_grid_cells = CellDataArray(
    ...     [[0, 1, 2, 3]],
    ...     coords=dict(cell_index=np.arange(1), vertex_index=np.arange(4)),
    ... )
    >>>
    >>> tet_grid_values = IndexedDataArray(
    ...     [1.0, 2.0, 3.0, 4.0], coords=dict(index=np.arange(4)),
    ... )
    >>>
    >>> tet_grid = TetrahedralGridDataset(
    ...     points=tet_grid_points,
    ...     cells=tet_grid_cells,
    ...     values=tet_grid_values,
    ... )
    """

    """ Fundametal parameters to set up based on grid dimensionality """

    @classmethod
    def _traingular_dataset_type(cls) -> type:
        """Corresponding class for triangular grid datasets. We need to know this when creating a triangular slice from a tetrahedral grid."""
        return TriangularGridDataset

    @classmethod
    def _point_dims(cls) -> pd.PositiveInt:
        """Dimensionality of stored grid point coordinates."""
        return 3

    @classmethod
    def _cell_num_vertices(cls) -> pd.PositiveInt:
        """Number of vertices in a cell."""
        return 4

    """ Convenience properties """

    @cached_property
    def _points_3d_array(self) -> Bound:
        """3D coordinates of grid points."""
        return self.points.data

    """ VTK interfacing """

    @classmethod
    @requires_vtk
    def _vtk_cell_type(cls):
        """VTK cell type to use in the VTK representation."""
        return vtk["mod"].VTK_TETRA

    @classmethod
    @requires_vtk
    def _from_vtk_obj(
        cls,
        vtk_obj,
        field=None,
        remove_degenerate_cells: bool = False,
        remove_unused_points: bool = False,
        values_type=IndexedDataArray,
        expect_complex: bool = False,
        ignore_invalid_cells: bool = False,
    ) -> TetrahedralGridDataset:
        """Initialize from a vtkUnstructuredGrid instance."""

        # read point, cells, and values info from a vtk instance
        cells_numpy = vtk["vtk_to_numpy"](vtk_obj.GetCells().GetConnectivityArray())
        points_numpy = vtk["vtk_to_numpy"](vtk_obj.GetPoints().GetData())
        values = cls._get_values_from_vtk(
            vtk_obj, len(points_numpy), field, values_type, expect_complex
        )

        # verify cell_types
        cells_types = vtk["vtk_to_numpy"](vtk_obj.GetCellTypesArray())
        invalid_cells = cells_types != cls._vtk_cell_type()
        if any(invalid_cells):
            if ignore_invalid_cells:
                cell_offsets = vtk["vtk_to_numpy"](vtk_obj.GetCells().GetOffsetsArray())
                valid_cell_offsets = cell_offsets[:-1][invalid_cells == 0]
                cells_numpy = cells_numpy[
                    np.ravel(
                        valid_cell_offsets[:, None]
                        + np.arange(cls._cell_num_vertices(), dtype=int)[None, :]
                    )
                ]
            else:
                raise DataError("Only tetrahedral 'vtkUnstructuredGrid' is currently supported")

        # pack point and cell information into Tidy3D arrays
        num_cells = len(cells_numpy) // cls._cell_num_vertices()
        cells_numpy = np.reshape(cells_numpy, (num_cells, cls._cell_num_vertices()))

        cells = CellDataArray(
            cells_numpy,
            coords={
                "cell_index": np.arange(num_cells),
                "vertex_index": np.arange(cls._cell_num_vertices()),
            },
        )

        points = PointDataArray(
            points_numpy,
            coords={"index": np.arange(len(points_numpy)), "axis": np.arange(cls._point_dims())},
        )

        if remove_degenerate_cells:
            cells = cls._remove_degenerate_cells(cells=cells)

        if remove_unused_points:
            points, values, cells = cls._remove_unused_points(
                points=points, values=values, cells=cells
            )

        return cls(points=points, cells=cells, values=values)

    """ Grid operations """

    @requires_vtk
    def plane_slice(self, axis: Axis, pos: float) -> TriangularGridDataset:
        """Slice data with a plane and return the resulting :class:`.TriangularGridDataset`.

        Parameters
        ----------
        axis : Axis
            The normal direction of the slicing plane.
        pos : float
            Position of the slicing plane along its normal direction.

        Returns
        -------
        TriangularGridDataset
            The resulting slice.
        """

        slice_vtk = self._plane_slice_raw(axis=axis, pos=pos)

        return self._traingular_dataset_type()._from_vtk_obj(
            slice_vtk,
            remove_degenerate_cells=True,
            remove_unused_points=True,
            field=self._values_coords_dict,
            values_type=self._values_type,
            expect_complex=self.is_complex,
        )

    @requires_vtk
    def line_slice(self, axis: Axis, pos: Coordinate) -> XrDataArray:
        """Slice data with a line and return the resulting xarray.DataArray.

        Parameters
        ----------
        axis : Axis
            The axis of the slicing line.
        pos : Tuple[float, float, float]
            Position of the slicing line.

        Returns
        -------
        xarray.DataArray
            The resulting slice.
        """

        bounds = self.bounds
        start = list(pos)
        end = list(pos)

        start[axis] = bounds[0][axis]
        end[axis] = bounds[1][axis]

        # create cutting plane
        line = vtk["mod"].vtkLineSource()
        line.SetPoint1(start)
        line.SetPoint2(end)
        line.SetResolution(1)

        # this should be done using vtkProbeLineFilter
        # but for some reason it crashes Python
        # so, we use a workaround:
        # 1) extract cells that are intersected by line (to speed up further slicing)
        # 2) do plane slice along first direction
        # 3) do second plane slice along second direction

        prober = vtk["mod"].vtkExtractCellsAlongPolyLine()
        prober.SetSourceConnection(line.GetOutputPort())
        prober.SetInputData(self._vtk_obj)
        prober.Update()

        extracted_cells_vtk = prober.GetOutput()

        if extracted_cells_vtk.GetNumberOfPoints() == 0:
            raise DataError("Slicing line does not intersect the unstructured grid.")

        extracted_cells = self._from_vtk_obj_internal(extracted_cells_vtk)

        tan_dims = [0, 1, 2]
        tan_dims.remove(axis)

        # first plane slice
        plane_slice = extracted_cells.plane_slice(axis=tan_dims[0], pos=pos[tan_dims[0]])
        # second plane slice
        line_slice = plane_slice.plane_slice(axis=tan_dims[1], pos=pos[tan_dims[1]])

        return line_slice

    """ Interpolation """

    def _interp_py(
        self,
        x: ArrayLike,
        y: ArrayLike,
        z: ArrayLike,
        fill_value: float,
        max_samples_per_step: int,
        max_cells_per_step: int,
        rel_tol: float,
    ) -> ArrayLike:
        """3D-specific function to interpolate data at provided x, y, and z
        using vectorized python implementation.

        Parameters
        ----------
        x : Union[float, ArrayLike]
            x-coordinates of sampling points.
        y : Union[float, ArrayLike]
            y-coordinates of sampling points.
        z : Union[float, ArrayLike]
            z-coordinates of sampling points.
        fill_value : float
            Value to use when filling points without interpolated values.
        max_samples_per_step : int
            Max number of points to interpolate at per iteration (used only if `use_vtk=False`).
            Using a higher number may speed up calculations but, at the same time, it increases
            RAM usage.
        max_cells_per_step : int
            Max number of cells to interpolate from per iteration (used only if `use_vtk=False`).
            Using a higher number may speed up calculations but, at the same time, it increases
            RAM usage.
        rel_tol : float
            Relative tolerance when determining whether a point belongs to a cell.

        Returns
        -------
        ArrayLike
            Interpolated data.
        """

        return self._interp_py_general(
            x=x,
            y=y,
            z=z,
            fill_value=fill_value,
            max_samples_per_step=max_samples_per_step,
            max_cells_per_step=max_cells_per_step,
            rel_tol=rel_tol,
            axis_ignore=None,
        )

    """ Data selection """

    @requires_vtk
    def sel(
        self,
        x: Union[float, ArrayLike] = None,
        y: Union[float, ArrayLike] = None,
        z: Union[float, ArrayLike] = None,
        method=None,
        **sel_kwargs,
    ) -> Union[TriangularGridDataset, XrDataArray]:
        """Extract/interpolate data along one or more spatial or non-spatial directions. Must provide at least one argument
        among 'x', 'y', 'z' or non-spatial dimensions through additional arguments. Along spatial dimensions a suitable slicing of
        grid is applied (plane slice, line slice, or interpolation). Selection along non-spatial dimensions is forwarded to
        .sel() xarray function. Parameter 'method' applies only to non-spatial dimensions.

        Parameters
        ----------
        x : Union[float, ArrayLike] = None
            x-coordinate of the slice.
        y : Union[float, ArrayLike] = None
            y-coordinate of the slice.
        z : Union[float, ArrayLike] = None
            z-coordinate of the slice.
        method: Literal[None, "nearest", "pad", "ffill", "backfill", "bfill"] = None
            Method to use in xarray sel() function.
        **sel_kwargs : dict
            Keyword arguments to pass to the xarray sel() function.

        Returns
        -------
        Union[TriangularGridDataset, xarray.DataArray]
            Extracted data.
        """

        xyz = [x, y, z]
        axes = [ind for ind, comp in enumerate(xyz) if comp is not None]

        num_provided = len(axes)

        if num_provided < 3 and any(not np.isscalar(comp) for comp in xyz if comp is not None):
            raise DataError(
                "Providing x, y, or z as array is only allowed for interpolation. That is, when all"
                " three x, y, and z are provided or method '.interp()' is used explicitly."
            )

        if num_provided == 0 and len(sel_kwargs) == 0:
            raise DataError(
                "Must provide at least one dimension to select along "
                "(available: {self._non_spatial_dims + list('xyz')})."
            )

        self_after_non_spatial_sel = self._non_spatial_sel(method=method, **sel_kwargs)

        if num_provided == 1:
            axis = axes[0]
            return self_after_non_spatial_sel.plane_slice(axis=axis, pos=xyz[axis])

        if num_provided == 2:
            axis = 3 - axes[0] - axes[1]
            xyz[axis] = 0
            return self_after_non_spatial_sel.line_slice(axis=axis, pos=xyz)

        if num_provided == 3:
            return self_after_non_spatial_sel.interp(x=x, y=y, z=z)

        return self_after_non_spatial_sel

    def get_cell_volumes(self):
        """Get the volumes associated to each cell in the grid"""
        v0 = self.points[self.cells.sel(vertex_index=0)]
        e01 = self.points[self.cells.sel(vertex_index=1)] - v0
        e02 = self.points[self.cells.sel(vertex_index=2)] - v0
        e03 = self.points[self.cells.sel(vertex_index=3)] - v0

        return np.abs(np.sum(np.cross(e01, e02) * e03, axis=1)) / 6



"""Imports from scattering matrix plugin."""

from __future__ import annotations

import warnings

from .component_modelers.modal import AbstractComponentModeler, ComponentModeler, ModalPortDataArray
from .component_modelers.terminal import TerminalComponentModeler
from .data.terminal import PortDataArray, TerminalPortDataArray
from .ports.coaxial_lumped import CoaxialLumpedPort
from .ports.modal import Port
from .ports.rectangular_lumped import LumpedPort
from .ports.wave import WavePort

# Instantiate on plugin import till we unite with toplevel
warnings.filterwarnings(
    "once",
    message="ℹ️ ⚠️ RF simulations are subject to new license requirements in the future. You have instantiated at least one RF-specific component.",
    category=FutureWarning,
)


__all__ = [
    "AbstractComponentModeler",
    "CoaxialLumpedPort",
    "ComponentModeler",
    "LumpedPort",
    "ModalPortDataArray",
    "Port",
    "PortDataArray",
    "TerminalComponentModeler",
    "TerminalPortDataArray",
    "WavePort",
]



"""Base model for Tidy3D components that are compatible with jax."""

from __future__ import annotations

import json
from typing import Any, Callable, Optional

import jax
import numpy as np
import pydantic.v1 as pd
from jax.tree_util import tree_flatten as jax_tree_flatten
from jax.tree_util import tree_unflatten as jax_tree_unflatten

from tidy3d.components.base import Tidy3dBaseModel

from .data.data_array import JAX_DATA_ARRAY_TAG, JaxDataArray

# end of the error message when a ``_validate_web_adjoint`` exception is raised
WEB_ADJOINT_MESSAGE = (
    "You can still run this simulation through "
    "'tidy3d.plugins.adjoint.web.run_local' or 'tidy3d.plugins.adjoint.web.run_local' "
    ", which are similar to 'run' / 'run_async', but "
    "perform the gradient postprocessing calculation locally after the simulation runs. "
    "Note that the postprocessing time can become "
    "quite long (several minutes or more) if these restrictions are exceeded. "
    "Furthermore, the local versions of 'adjoint' require downloading field data "
    "inside of the 'input_structures', which can greatly increase the size of data "
    "needing to be downloaded."
)


class JaxObject(Tidy3dBaseModel):
    """Abstract class that makes a :class:`.Tidy3dBaseModel` jax-compatible through inheritance."""

    _tidy3d_class = Tidy3dBaseModel

    """Shortcut to get names of fields with certain properties."""

    @classmethod
    def _get_field_names(cls, field_key: str) -> list[str]:
        """Get all fields where ``field_key`` defined in the ``pydantic.Field``."""
        fields = []
        for field_name, model_field in cls.__fields__.items():
            field_value = model_field.field_info.extra.get(field_key)
            if field_value:
                fields.append(field_name)
        return fields

    @classmethod
    def get_jax_field_names(cls) -> list[str]:
        """Returns list of field names where ``jax_field=True``."""
        return cls._get_field_names("jax_field")

    @classmethod
    def get_jax_leaf_names(cls) -> list[str]:
        """Returns list of field names where ``stores_jax_for`` defined."""
        return cls._get_field_names("stores_jax_for")

    @classmethod
    def get_jax_field_names_all(cls) -> list[str]:
        """Returns list of field names where ``jax_field=True`` or ``stores_jax_for`` defined."""
        jax_field_names = cls.get_jax_field_names()
        jax_leaf_names = cls.get_jax_leaf_names()
        return list(set(jax_field_names + jax_leaf_names))

    @property
    def jax_fields(self) -> dict:
        """Get dictionary of ``jax`` fields."""

        # TODO: don't use getattr, define this dictionary better
        jax_field_names = self.get_jax_field_names()
        return {key: getattr(self, key) for key in jax_field_names}

    def _validate_web_adjoint(self) -> None:
        """Run validators for this component, only if using ``tda.web.run()``."""

    """Methods needed for jax to register arbitrary classes."""

    def tree_flatten(self) -> tuple[list, dict]:
        """How to flatten a :class:`.JaxObject` instance into a ``pytree``."""
        children = []
        aux_data = self.dict()

        for field_name in self.get_jax_field_names_all():
            field = getattr(self, field_name)
            sub_children, sub_aux_data = jax_tree_flatten(field)
            children.append(sub_children)
            aux_data[field_name] = sub_aux_data

        def fix_numpy(value: Any) -> Any:
            """Recursively convert any ``numpy`` array in the value to nested list."""
            if isinstance(value, (tuple, list)):
                return [fix_numpy(val) for val in value]
            if isinstance(value, np.ndarray):
                return value.tolist()
            if isinstance(value, dict):
                return {key: fix_numpy(val) for key, val in value.items()}
            return value

        aux_data = fix_numpy(aux_data)

        return children, aux_data

    @classmethod
    def tree_unflatten(cls, aux_data: dict, children: list) -> JaxObject:
        """How to unflatten a ``pytree`` into a :class:`.JaxObject` instance."""
        self_dict = aux_data.copy()
        for field_name, sub_children in zip(cls.get_jax_field_names_all(), children):
            sub_aux_data = aux_data[field_name]
            field = jax_tree_unflatten(sub_aux_data, sub_children)
            self_dict[field_name] = field

        return cls.parse_obj(self_dict)

    """Type conversion helpers."""

    def to_tidy3d(self: JaxObject) -> Tidy3dBaseModel:
        """Convert :class:`.JaxObject` instance to :class:`.Tidy3dBaseModel` instance."""

        self_dict = self.dict(exclude=self.exclude_fields_leafs_only)

        for key in self.get_jax_field_names():
            sub_field = self.jax_fields[key]

            # TODO: simplify this logic
            if isinstance(sub_field, (tuple, list)):
                self_dict[key] = [x.to_tidy3d() for x in sub_field]
            else:
                self_dict[key] = sub_field.to_tidy3d()
            # end TODO

        return self._tidy3d_class.parse_obj(self_dict)

    @classmethod
    def from_tidy3d(cls, tidy3d_obj: Tidy3dBaseModel) -> JaxObject:
        """Convert :class:`.Tidy3dBaseModel` instance to :class:`.JaxObject`."""
        obj_dict = tidy3d_obj.dict(exclude={"type"})

        for key in cls.get_jax_field_names():
            sub_field_type = cls.__fields__[key].type_
            tidy3d_sub_field = getattr(tidy3d_obj, key)

            # TODO: simplify this logic
            if isinstance(tidy3d_sub_field, (tuple, list)):
                obj_dict[key] = [sub_field_type.from_tidy3d(x) for x in tidy3d_sub_field]
            else:
                obj_dict[key] = sub_field_type.from_tidy3d(tidy3d_sub_field)
            # end TODO

        return cls.parse_obj(obj_dict)

    @property
    def exclude_fields_leafs_only(self) -> set:
        """Fields to exclude from ``self.dict()``, ``"type"`` and all ``jax`` leafs."""
        return {"type", *self.get_jax_leaf_names()}

    """Accounting with jax and regular fields."""

    @pd.root_validator(pre=True)
    def handle_jax_kwargs(cls, values: dict) -> dict:
        """Pass jax inputs to the jax fields and pass untraced values to the regular fields."""

        # for all jax-traced fields
        for jax_name in cls.get_jax_leaf_names():
            # if a value was passed to the object for the regular field
            orig_name = cls.__fields__[jax_name].field_info.extra.get("stores_jax_for")
            val = values.get(orig_name)
            if val is not None:
                # try adding the sanitized (no trace) version to the regular field
                try:
                    values[orig_name] = jax.lax.stop_gradient(val)

                # if it doesnt work, just pass the raw value (necessary to handle inf strings)
                except TypeError:
                    values[orig_name] = val

                # if the jax name was not specified directly, use the original traced value
                if jax_name not in values:
                    values[jax_name] = val

        return values

    @pd.root_validator(pre=True)
    def handle_array_jax_leafs(cls, values: dict) -> dict:
        """Convert jax_leafs that are passed as numpy arrays."""
        for jax_name in cls.get_jax_leaf_names():
            val = values.get(jax_name)
            if isinstance(val, np.ndarray):
                values[jax_name] = val.tolist()
        return values

    """ IO """

    # TODO: replace with JaxObject json encoder

    def _json(self, *args, **kwargs) -> str:
        """Overwritten method to get the json string to store in the files."""

        json_string_og = super()._json(*args, **kwargs)
        json_dict = json.loads(json_string_og)

        def strip_data_array(val: Any) -> Any:
            """Recursively strip any elements with type "JaxDataArray", replace with tag."""

            if isinstance(val, dict):
                if "type" in val and val["type"] == "JaxDataArray":
                    return JAX_DATA_ARRAY_TAG
                return {k: strip_data_array(v) for k, v in val.items()}

            if isinstance(val, (tuple, list)):
                return [strip_data_array(v) for v in val]

            return val

        json_dict = strip_data_array(json_dict)
        return json.dumps(json_dict)

    # TODO: replace with implementing these in DataArray

    def to_hdf5(self, fname: str, custom_encoders: Optional[list[Callable]] = None) -> None:
        """Exports :class:`JaxObject` instance to .hdf5 file.

        Parameters
        ----------
        fname : str
            Full path to the .hdf5 file to save the :class:`JaxObject` to.
        custom_encoders : List[Callable]
            List of functions accepting (fname: str, group_path: str, value: Any) that take
            the ``value`` supplied and write it to the hdf5 ``fname`` at ``group_path``.

        Example
        -------
        >>> simulation.to_hdf5(fname='folder/sim.hdf5') # doctest: +SKIP
        """

        def data_array_encoder(fname: str, group_path: str, value: Any) -> None:
            """Custom encoder to convert the JaxDataArray dict to an instance."""
            if isinstance(value, dict) and "type" in value and value["type"] == "JaxDataArray":
                data_array = JaxDataArray(values=value["values"], coords=value["coords"])
                data_array.to_hdf5(fname=fname, group_path=group_path)

        if custom_encoders is None:
            custom_encoders = []

        custom_encoders += [data_array_encoder]

        return super().to_hdf5(fname=fname, custom_encoders=custom_encoders)

    @classmethod
    def dict_from_hdf5(
        cls, fname: str, group_path: str = "", custom_decoders: Optional[list[Callable]] = None
    ) -> dict:
        """Loads a dictionary containing the model contents from a .hdf5 file.

        Parameters
        ----------
        fname : str
            Full path to the .hdf5 file to load the :class:`JaxObject` from.
        group_path : str, optional
            Path to a group inside the file to selectively load a sub-element of the model only.
        custom_decoders : List[Callable]
            List of functions accepting
            (fname: str, group_path: str, model_dict: dict, key: str, value: Any) that store the
            value in the model dict after a custom decoding.

        Returns
        -------
        dict
            Dictionary containing the model.

        Example
        -------
        >>> sim_dict = Simulation.dict_from_hdf5(fname='folder/sim.hdf5') # doctest: +SKIP
        """

        def data_array_decoder(
            fname: str, group_path: str, model_dict: dict, key: str, value: Any
        ) -> None:
            """Custom decoder to grab JaxDataArray from file and save it in model_dict."""

            # write the path to the element of the json dict where the data_array should be
            if isinstance(value, str) and value == JAX_DATA_ARRAY_TAG:
                jax_data_array = JaxDataArray.from_hdf5(fname=fname, group_path=group_path)
                model_dict[key] = jax_data_array

        if custom_decoders is None:
            custom_decoders = []

        custom_decoders += [data_array_decoder]

        return super().dict_from_hdf5(
            fname=fname, group_path=group_path, custom_decoders=custom_decoders
        )



"""Utilities for 2D geometry manipulation."""

from __future__ import annotations

from math import isclose

import numpy as np
import shapely

from tidy3d.components.geometry.base import Box, ClipOperation, Geometry
from tidy3d.components.geometry.polyslab import _MIN_POLYGON_AREA, PolySlab
from tidy3d.components.grid.grid import Grid
from tidy3d.components.scene import Scene
from tidy3d.components.structure import Structure
from tidy3d.components.types import Axis
from tidy3d.constants import fp_eps, inf


def increment_float(val: float, sign) -> float:
    """Applies a small positive or negative shift as though `val` is a 32bit float
    using numpy.nextafter, but additionally handles some corner cases.
    """
    # Infinity is left unchanged
    if val == inf or val == -inf:
        return val

    if sign >= 0:
        sign = 1
    else:
        sign = -1

    # Avoid small increments within subnormal values
    if np.abs(val) <= np.finfo(np.float32).tiny:
        return val + sign * np.finfo(np.float32).tiny

    # Numpy seems to skip over the increment from -0.0 and +0.0
    # which is different from c++
    val_inc = np.nextafter(val, sign * inf, dtype=np.float32)

    return np.float32(val_inc)


def snap_coordinate_to_grid(grid: Grid, center: float, axis: Axis) -> float:
    """2D materials are snapped to grid along their normal axis"""
    new_centers = grid.boundaries.to_list[axis]
    new_center = new_centers[np.argmin(abs(new_centers - center))]
    return new_center


def get_bounds(geom: Geometry, axis: Axis) -> tuple[float, float]:
    """Get the bounds of a geometry in the axis direction."""
    return (geom.bounds[0][axis], geom.bounds[1][axis])


def get_thickened_geom(geom: Geometry, axis: Axis):
    """Helper to return a slightly thickened version of a planar geometry."""
    center = get_bounds(geom, axis)[0]
    neg_thickness = increment_float(center, -1.0)
    pos_thickness = increment_float(center, 1.0)
    return geom._update_from_bounds(bounds=(neg_thickness, pos_thickness), axis=axis)


def get_neighbors(
    geom: Geometry,
    axis: Axis,
    structures: list[Structure],
):
    """Find the neighboring structures and return the tested positions above and below."""
    center = get_bounds(geom, axis)[0]
    check_delta = [
        increment_float(center, -1.0) - center,
        increment_float(center, 1.0) - center,
    ]

    neighbors_below = []
    neighbors_above = []
    for _, position in enumerate(check_delta):
        geom_shifted = geom._update_from_bounds(
            bounds=(center + position, center + position), axis=axis
        )

        # to prevent false positives due to 2D materials touching different materials
        # along their sides, shrink the bounds along the tangential directions by
        # a tiny bit before checking for intersections
        bounds = [list(i) for i in geom_shifted.bounds]
        _, tan_dirs = Geometry.pop_axis([0, 1, 2], axis=axis)
        for dim in tan_dirs:
            # Don't shrink if the width is already close to 0
            if not isclose(bounds[0][dim], bounds[1][dim], rel_tol=2 * fp_eps):
                bounds[0][dim] = increment_float(bounds[0][dim], 1.0)
                bounds[1][dim] = increment_float(bounds[1][dim], -1.0)

        structures_side = Scene.intersecting_structures(Box.from_bounds(*bounds), structures)

        if position < 0:
            neighbors_below += list(structures_side)
        else:
            neighbors_above += list(structures_side)

    return neighbors_below, neighbors_above, check_delta


def subdivide(
    geom: Geometry, structures: list[Structure]
) -> list[tuple[Geometry, Structure, Structure]]:
    """Subdivide geometry associated with a :class:`.Medium2D` into partitions
    that each have a homogeneous substrate / superstrate. Partitions are computed
    using ``shapely`` boolean operations on polygons.

    Parameters
    ----------
    geom : Geometry
        A 2D geometry associated with the :class:`.Medium2D`.
    structures : List[Structure]
        List of structures that are checked for intersection with ``geom``.

    Returns
    -------
    List[Tuple[Geometry, Structure, Structure]]
        List of the created partitions. Each element of the list represents a partition of the 2D geometry,
        which includes the newly created structures below and above.

    """

    def shapely_to_polyslab(polygon: shapely.Polygon, axis: Axis, center: float) -> PolySlab:
        xx, yy = polygon.exterior.coords.xy
        vertices = list(zip(xx, yy))
        return PolySlab(slab_bounds=(center, center), vertices=vertices, axis=axis)

    def to_multipolygon(shapely_geometry) -> shapely.MultiPolygon:
        return shapely.MultiPolygon(ClipOperation.to_polygon_list(shapely_geometry))

    axis = geom._normal_2dmaterial
    # Find neighbors and the small offset they were found at
    neighbors_below, neighbors_above, check_delta = get_neighbors(
        geom=geom, axis=axis, structures=structures
    )

    # Compute the plane of intersection
    center = get_bounds(geom, axis)[0]
    coord = "xyz"[axis]
    plane = {coord: center}

    # Convert input geometry into MultiPolygon shapely geometry and track the original structure that references the media properties
    geom_shapely = Geometry.evaluate_inf_shape(shapely.union_all(geom.intersections_plane(**plane)))

    plane[coord] = center + check_delta[1]
    above_shapely = [
        (
            Geometry.evaluate_inf_shape(
                shapely.union_all(structure.geometry.intersections_plane(**plane))
            ),
            structure,
        )
        for structure in neighbors_above
    ]

    plane[coord] = center + check_delta[0]
    below_shapely = [
        [
            Geometry.evaluate_inf_shape(
                shapely.union_all(structure.geometry.intersections_plane(**plane))
            ),
            structure,
        ]
        for structure in neighbors_below
    ]

    # First find the intersections of 2d material with all structures above in reverse order
    above_intersections = []
    for mp_structure in reversed(above_shapely):
        # If the 2D structure overlaps completely with all previously tested structures above then there is no more work to do
        if not geom_shapely:
            break

        intersection_res = shapely.intersection(geom_shapely, mp_structure[0])
        intersection_mp = to_multipolygon(intersection_res)
        difference_res = shapely.difference(geom_shapely, mp_structure[0])
        geom_shapely = to_multipolygon(difference_res)

        if intersection_mp:
            above_intersections.append((intersection_mp, mp_structure[1]))
    above_intersections.reverse()

    # Next find intersections of previous result with all structures below
    # List that stores a tuple of a MultiPolygon, the adjacent structure below, and the adjacent structure above
    both_intersections = []
    # Similar to above, but keep track of both differences of the previous result and the below polygons for faster termination
    for mp_structure_above in reversed(above_intersections):
        above_intersection = mp_structure_above[0]
        for mp_structure_below in reversed(below_shapely):
            # Possible to finish loops early
            if not above_intersection:
                break
            if not mp_structure_below[0]:
                continue

            intersection_res = shapely.intersection(above_intersection, mp_structure_below[0])
            intersection_mp = to_multipolygon(intersection_res)
            above_difference = to_multipolygon(
                shapely.difference(above_intersection, mp_structure_below[0])
            )
            below_difference = to_multipolygon(
                shapely.difference(mp_structure_below[0], above_intersection)
            )
            # Update polygons by subtracting the intersecting parts
            mp_structure_below[0] = below_difference
            above_intersection = above_difference
            if intersection_mp:
                both_intersections.append(
                    (intersection_mp, mp_structure_below[1], mp_structure_above[1])
                )
    both_intersections.reverse()

    # If there turns out to be only one substrate/superstrate combo, then return the original geometry
    if len(both_intersections) == 1:
        return [(geom, both_intersections[0][1], both_intersections[0][2])]

    # Flatten into array of only polygons and adjacent structures
    # The geometry produced should all be MultiPolygons
    final_polygons = []
    for element in both_intersections:
        for polygon in element[0].geoms:
            final_polygons.append((polygon, element[1], element[2]))

    # Create polyslab from subdivided geometry, while filtering out any
    # polygons with very small areas
    polyslab_result = [
        (shapely_to_polyslab(element[0], axis, center), element[1], element[2])
        for element in final_polygons
        if element[0].area >= _MIN_POLYGON_AREA
    ]

    return polyslab_result



"""Defines a jax-compatible SimulationData."""

from __future__ import annotations

from typing import Optional, Union

import numpy as np
import pydantic.v1 as pd
import xarray as xr
from jax.tree_util import register_pytree_node_class

from tidy3d.components.data.monitor_data import FieldData, MonitorDataType, PermittivityData
from tidy3d.components.data.sim_data import SimulationData
from tidy3d.components.source.current import PointDipole
from tidy3d.components.source.time import GaussianPulse
from tidy3d.log import log
from tidy3d.plugins.adjoint.components.base import JaxObject
from tidy3d.plugins.adjoint.components.simulation import JaxInfo, JaxSimulation

from .monitor_data import JAX_MONITOR_DATA_MAP, JaxMonitorDataType


@register_pytree_node_class
class JaxSimulationData(SimulationData, JaxObject):
    """A :class:`.SimulationData` registered with jax."""

    output_data: tuple[JaxMonitorDataType, ...] = pd.Field(
        (),
        title="Jax Data",
        description="Tuple of Jax-compatible data associated with output monitors.",
        jax_field=True,
    )

    grad_data: tuple[FieldData, ...] = pd.Field(
        (),
        title="Gradient Field Data",
        description="Tuple of monitor data storing fields associated with the input structures.",
    )

    grad_eps_data: tuple[PermittivityData, ...] = pd.Field(
        (),
        title="Gradient Permittivity Data",
        description="Tuple of monitor data storing epsilon associated with the input structures.",
    )

    simulation: JaxSimulation = pd.Field(
        ...,
        title="Simulation",
        description="The jax-compatible simulation corresponding to the data.",
    )

    task_id: str = pd.Field(
        None,
        title="Task ID",
        description="Optional field storing the task_id for the original JaxSimulation.",
    )

    def get_poynting_vector(self, field_monitor_name: str) -> xr.Dataset:
        """return ``xarray.Dataset`` of the Poynting vector at Yee cell centers.

        Calculated values represent the instantaneous Poynting vector for time-domain fields and the
        complex vector for frequency-domain: ``S = 1/2 E × conj(H)``.

        Only the available components are returned, e.g., if the indicated monitor doesn't include
        field component `"Ex"`, then `"Sy"` and `"Sz"` will not be calculated.

        Parameters
        ----------
        field_monitor_name : str
            Name of field monitor used in the original :class:`Simulation`.

        Returns
        -------
        xarray.DataArray
            DataArray containing the Poynting vector calculated based on the field components
            colocated at the center locations of the Yee grid.
        """

        if field_monitor_name in self.output_monitor_data:
            raise NotImplementedError(
                "Adjoint support for differentiation with respect to Poynting vector not available."
            )

        return super().get_poynting_vector(field_monitor_name)

    @property
    def grad_data_symmetry(self) -> tuple[FieldData, ...]:
        """``self.grad_data`` but with ``symmetry_expanded_copy`` applied."""
        return tuple(data.symmetry_expanded_copy for data in self.grad_data)

    @property
    def grad_eps_data_symmetry(self) -> tuple[FieldData, ...]:
        """``self.grad_eps_data`` but with ``symmetry_expanded_copy`` applied."""
        return tuple(data.symmetry_expanded_copy for data in self.grad_eps_data)

    @property
    def output_monitor_data(self) -> dict[str, JaxMonitorDataType]:
        """Dictionary of ``.output_data`` monitor ``.name`` to the corresponding data."""
        return {monitor_data.monitor.name: monitor_data for monitor_data in self.output_data}

    @property
    def monitor_data(self) -> dict[str, Union[JaxMonitorDataType, MonitorDataType]]:
        """Dictionary of ``.output_data`` monitor ``.name`` to the corresponding data."""
        reg_mnt_data = {monitor_data.monitor.name: monitor_data for monitor_data in self.data}
        reg_mnt_data.update(self.output_monitor_data)
        return reg_mnt_data

    @staticmethod
    def split_data(
        mnt_data: list[MonitorDataType], jax_info: JaxInfo
    ) -> dict[str, list[MonitorDataType]]:
        """Split list of monitor data into data, output_data, grad_data, and grad_eps_data."""
        # Get information needed to split the full data list
        len_output_data = jax_info.num_output_monitors
        len_grad_data = jax_info.num_grad_monitors
        len_grad_eps_data = jax_info.num_grad_eps_monitors
        len_data = len(mnt_data) - len_output_data - len_grad_data - len_grad_eps_data

        # split the data list into regular data, output_data, and grad_data
        all_data = list(mnt_data)
        data = all_data[:len_data]
        output_data = all_data[len_data : len_data + len_output_data]
        grad_data = all_data[
            len_data + len_output_data : len_data + len_output_data + len_grad_data
        ]
        grad_eps_data = all_data[len_data + len_output_data + len_grad_data :]

        return {
            "data": data,
            "output_data": output_data,
            "grad_data": grad_data,
            "grad_eps_data": grad_eps_data,
        }

    @classmethod
    def from_sim_data(
        cls, sim_data: SimulationData, jax_info: JaxInfo, task_id: Optional[str] = None
    ) -> JaxSimulationData:
        """Construct a :class:`.JaxSimulationData` instance from a :class:`.SimulationData`."""

        self_dict = sim_data.dict(exclude={"type", "simulation", "data"})

        # convert the simulation to JaxSimulation
        jax_sim = JaxSimulation.from_simulation(simulation=sim_data.simulation, jax_info=jax_info)

        # construct JaxSimulationData with no data (yet)
        self_dict["simulation"] = jax_sim
        self_dict["data"] = ()

        data_dict = cls.split_data(mnt_data=sim_data.data, jax_info=jax_info)

        # convert the output data to the proper jax type
        output_data_list = []
        for mnt_data in data_dict["output_data"]:
            mnt_data_type_str = type(mnt_data)
            if mnt_data_type_str not in JAX_MONITOR_DATA_MAP:
                raise KeyError(
                    f"MonitorData type '{mnt_data_type_str}' "
                    "not currently supported by adjoint plugin."
                )
            mnt_data_type = JAX_MONITOR_DATA_MAP[mnt_data_type_str]
            jax_mnt_data = mnt_data_type.from_monitor_data(mnt_data)
            output_data_list.append(jax_mnt_data)
        data_dict["output_data"] = output_data_list
        self_dict.update(data_dict)
        self_dict.update({"task_id": task_id})

        return cls.parse_obj(self_dict)

    @classmethod
    def split_fwd_sim_data(
        cls, sim_data: SimulationData, jax_info: JaxInfo
    ) -> tuple[SimulationData, SimulationData]:
        """Split a :class:`.SimulationData` into two parts, containing user and gradient data."""

        sim = sim_data.simulation

        data_dict = cls.split_data(mnt_data=sim_data.data, jax_info=jax_info)
        user_data = data_dict["data"] + data_dict["output_data"]
        adjoint_data = data_dict["grad_data"] + data_dict["grad_eps_data"]

        mnt_dict = JaxSimulation.split_monitors(
            monitors=sim_data.simulation.monitors, jax_info=jax_info
        )
        user_mnts = mnt_dict["monitors"] + mnt_dict["output_monitors"]
        adjoint_mnts = mnt_dict["grad_monitors"] + mnt_dict["grad_eps_monitors"]

        user_sim = sim.updated_copy(monitors=user_mnts)
        adjoint_sim = sim.updated_copy(monitors=adjoint_mnts)

        user_sim_data = sim_data.updated_copy(data=user_data, simulation=user_sim)
        adjoint_sim_data = sim_data.updated_copy(data=adjoint_data, simulation=adjoint_sim)

        return user_sim_data, adjoint_sim_data

    def make_adjoint_simulation(self, fwidth: float, run_time: float) -> JaxSimulation:
        """Make an adjoint simulation out of the data provided (generally, the vjp sim data)."""

        sim_fwd = self.simulation

        # grab boundary conditions with flipped bloch vectors (for adjoint)
        bc_adj = sim_fwd.boundary_spec.flipped_bloch_vecs

        # add all adjoint sources and boundary conditions (at same time for BC validators to work)
        adj_srcs = []
        for mnt_data_vjp in self.output_data:
            for adj_source in mnt_data_vjp.to_adjoint_sources(fwidth=fwidth):
                adj_srcs.append(adj_source)

        # in this case (no adjoint sources) give it an "empty" source
        if not adj_srcs:
            log.warning(
                "No adjoint sources, making a mock source with amplitude = 0. "
                "All gradients will be zero for anything depending on this simulation's data. "
                "This comes up when a simulation's data contributes to the value of an objective "
                "function but the contribution from each member of the data is 0. "
                "If this is intended (eg. if using 'jnp.max()' of several simulation results), "
                "please ignore. Otherwise, this can suggest a mistake in your objective function."
            )

            # set a zero-amplitude source
            adj_srcs.append(
                PointDipole(
                    center=sim_fwd.center,
                    polarization="Ez",
                    source_time=GaussianPulse(
                        freq0=sim_fwd.freqs_adjoint[0],
                        fwidth=sim_fwd._fwidth_adjoint,
                        amplitude=0.0,
                    ),
                )
            )

            # set a very short run time relative to the fwidth
            run_time = 2 / fwidth

        update_dict = {
            "boundary_spec": bc_adj,
            "sources": adj_srcs,
            "monitors": (),
            "output_monitors": (),
            "run_time": run_time,
            "normalize_index": None,  # normalize later, frequency-by-frequency
        }

        update_dict.update(
            sim_fwd.get_grad_monitors(
                input_structures=sim_fwd.input_structures,
                freqs_adjoint=sim_fwd.freqs_adjoint,
                include_eps_mnts=False,
            )
        )

        # set the ADJ grid spec wavelength to the FWD wavelength (for same meshing)
        grid_spec_fwd = sim_fwd.grid_spec
        if len(sim_fwd.sources) and grid_spec_fwd.wavelength is None:
            wavelength_fwd = grid_spec_fwd.wavelength_from_sources(sim_fwd.sources)
            grid_spec_adj = grid_spec_fwd.updated_copy(wavelength=wavelength_fwd)
            update_dict.update({"grid_spec": grid_spec_adj})

        return sim_fwd.updated_copy(**update_dict)

    def normalize_adjoint_fields(self) -> JaxSimulationData:
        """Make copy of jax_sim_data with grad_data (fields) normalized by adjoint sources."""

        grad_data_norm = []
        for field_data in self.grad_data:
            field_components_norm = {}
            for field_name, field_component in field_data.field_components.items():
                freqs = field_component.coords["f"]
                norm_factor_f = np.zeros(len(freqs), dtype=complex)
                for i, freq in enumerate(freqs):
                    freq = float(freq)
                    for source_index, source in enumerate(self.simulation.sources):
                        if source.source_time.freq0 == freq and source.source_time.amplitude > 0:
                            spectrum_fn = self.source_spectrum(source_index)
                            norm_factor_f[i] = complex(spectrum_fn([freq])[0])

                norm_factor_f_darr = xr.DataArray(norm_factor_f, coords={"f": freqs})
                field_component_norm = field_component / norm_factor_f_darr
                field_components_norm[field_name] = field_component_norm

            field_data_norm = field_data.updated_copy(**field_components_norm)
            grad_data_norm.append(field_data_norm)

        return self.updated_copy(grad_data=grad_data_norm)



"""connect util for webapi."""

from __future__ import annotations

import time
from functools import wraps

from requests import ReadTimeout
from requests.exceptions import ConnectionError as ConnErr
from requests.exceptions import JSONDecodeError
from urllib3.exceptions import NewConnectionError

from tidy3d.exceptions import WebError
from tidy3d.log import log

# number of seconds to keep re-trying connection before erroring
CONNECTION_RETRY_TIME = 180
# time between checking task status
REFRESH_TIME = 2


def wait_for_connection(decorated_fn=None, wait_time_sec: float = CONNECTION_RETRY_TIME):
    """Causes function to ignore connection errors and retry for ``wait_time_sec`` secs."""

    def decorator(web_fn):
        """Decorator returned by @wait_for_connection()"""

        @wraps(web_fn)
        def web_fn_wrapped(*args, **kwargs):
            """Function to return including connection waiting."""
            time_start = time.time()
            warned_previously = False

            while (time.time() - time_start) < wait_time_sec:
                try:
                    return web_fn(*args, **kwargs)
                except (ConnErr, ConnectionError, NewConnectionError, ReadTimeout, JSONDecodeError):
                    if not warned_previously:
                        log.warning(f"No connection: Retrying for {wait_time_sec} seconds.")
                        warned_previously = True
                    time.sleep(REFRESH_TIME)

            raise WebError("No internet connection: giving up on connection waiting.")

        return web_fn_wrapped

    if decorated_fn:
        return decorator(decorated_fn)

    return decorator


def get_time_steps_str(time_steps) -> str:
    """get_time_steps_str"""
    if time_steps < 1000:
        time_steps_str = f"{time_steps}"
    elif 1000 <= time_steps < 1000 * 1000:
        time_steps_str = f"{time_steps / 1000}K"
    else:
        time_steps_str = f"{time_steps / 1000 / 1000}M"
    return time_steps_str


def get_grid_points_str(grid_points) -> str:
    """get_grid_points_str"""
    if grid_points < 1000:
        grid_points_str = f"{grid_points}"
    elif 1000 <= grid_points < 1000 * 1000:
        grid_points_str = f"{grid_points / 1000}K"
    else:
        grid_points_str = f"{grid_points / 1000 / 1000}M"
    return grid_points_str



from __future__ import annotations

from typing import Callable, Optional, Union

import pydantic.v1 as pd
from numpy.typing import NDArray

from tidy3d.components.base import Tidy3dBaseModel
from tidy3d.plugins.autograd.constants import BETA_DEFAULT, ETA_DEFAULT
from tidy3d.plugins.autograd.types import KernelType, PaddingType

from .filters import make_filter
from .projections import tanh_projection


class FilterAndProject(Tidy3dBaseModel):
    """A class that combines filtering and projection operations."""

    radius: Union[float, tuple[float, ...]] = pd.Field(
        ..., title="Radius", description="The radius of the kernel."
    )
    dl: Union[float, tuple[float, ...]] = pd.Field(
        ..., title="Grid Spacing", description="The grid spacing."
    )
    size_px: Union[int, tuple[int, ...]] = pd.Field(
        None, title="Size in Pixels", description="The size of the kernel in pixels."
    )
    beta: pd.NonNegativeFloat = pd.Field(
        BETA_DEFAULT, title="Beta", description="The beta parameter for the tanh projection."
    )
    eta: pd.NonNegativeFloat = pd.Field(
        ETA_DEFAULT, title="Eta", description="The eta parameter for the tanh projection."
    )
    filter_type: KernelType = pd.Field(
        "conic", title="Filter Type", description="The type of filter to create."
    )
    padding: PaddingType = pd.Field(
        "reflect", title="Padding", description="The padding mode to use."
    )

    def __call__(
        self, array: NDArray, beta: Optional[float] = None, eta: Optional[float] = None
    ) -> NDArray:
        """Apply the filter and projection to an input array.

        Parameters
        ----------
        array : np.ndarray
            The input array to filter and project.
        beta : float = None
            The beta parameter for the tanh projection. If None, uses the instance's beta.
        eta : float = None
            The eta parameter for the tanh projection. If None, uses the instance's eta.

        Returns
        -------
        np.ndarray
            The filtered and projected array.
        """
        filter_instance = make_filter(
            radius=self.radius,
            dl=self.dl,
            size_px=self.size_px,
            filter_type=self.filter_type,
            padding=self.padding,
        )
        filtered = filter_instance(array)
        beta = beta if beta is not None else self.beta
        eta = eta if eta is not None else self.eta
        projected = tanh_projection(filtered, beta, eta)
        return projected


def make_filter_and_project(
    radius: Optional[Union[float, tuple[float, ...]]] = None,
    dl: Optional[Union[float, tuple[float, ...]]] = None,
    *,
    size_px: Optional[Union[int, tuple[int, ...]]] = None,
    beta: float = BETA_DEFAULT,
    eta: float = ETA_DEFAULT,
    filter_type: KernelType = "conic",
    padding: PaddingType = "reflect",
) -> Callable:
    """Create a function that filters and projects an array.

    See Also
    --------
    :func:`~parametrizations.FilterAndProject`.
    """
    return FilterAndProject(
        radius=radius,
        dl=dl,
        size_px=size_px,
        beta=beta,
        eta=eta,
        filter_type=filter_type,
        padding=padding,
    )



"""Class and custom data array for representing a scattering-matrix port, which is defined by a pair of terminals."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Optional, Union

import pydantic.v1 as pd

from tidy3d.components.base import Tidy3dBaseModel, cached_property
from tidy3d.components.data.data_array import FreqDataArray
from tidy3d.components.data.sim_data import SimulationData
from tidy3d.components.grid.grid import Grid
from tidy3d.components.monitor import FieldMonitor, ModeMonitor
from tidy3d.components.source.base import Source
from tidy3d.components.source.time import GaussianPulse
from tidy3d.components.types import FreqArray
from tidy3d.log import log


class AbstractTerminalPort(Tidy3dBaseModel, ABC):
    """Class representing a single terminal-based port. All terminal ports must provide methods
    for computing voltage and current. These quantities represent the voltage between the
    terminals, and the current flowing from one terminal into the other.
    """

    name: str = pd.Field(
        ...,
        title="Name",
        description="Unique name for the port.",
        min_length=1,
    )

    @cached_property
    @abstractmethod
    def injection_axis(self):
        """Injection axis of the port."""

    @abstractmethod
    def to_source(
        self, source_time: GaussianPulse, snap_center: Optional[float] = None, grid: Grid = None
    ) -> Source:
        """Create a current source from a terminal-based port."""

    def to_field_monitors(
        self, freqs: FreqArray, snap_center: Optional[float] = None, grid: Grid = None
    ) -> Union[list[FieldMonitor], list[ModeMonitor]]:
        """DEPRECATED: Monitors used to compute the port voltage and current."""
        log.warning(
            "'to_field_monitors' method name is deprecated and will be removed in the future. Please use "
            "'to_monitors' for the same effect."
        )
        return self.to_monitors(freqs=freqs, snap_center=snap_center, grid=grid)

    @abstractmethod
    def to_monitors(
        self, freqs: FreqArray, snap_center: Optional[float] = None, grid: Grid = None
    ) -> Union[list[FieldMonitor], list[ModeMonitor]]:
        """Monitors used to compute the port voltage and current."""

    @abstractmethod
    def compute_voltage(self, sim_data: SimulationData) -> FreqDataArray:
        """Helper to compute voltage across the port."""

    @abstractmethod
    def compute_current(self, sim_data: SimulationData) -> FreqDataArray:
        """Helper to compute current flowing into the port."""

    @pd.root_validator(pre=False)
    def _warn_rf_license(cls, values):
        log.warning(
            "ℹ️ ⚠️ RF simulations are subject to new license requirements in the future. You have instantiated at least one RF-specific component.",
            log_once=True,
        )
        return values



"""Find corners of structures on a 2D plane."""

from __future__ import annotations

from typing import Any, Literal, Optional

import numpy as np
import pydantic.v1 as pd

from tidy3d.components.base import Tidy3dBaseModel, cached_property
from tidy3d.components.geometry.base import Box, ClipOperation
from tidy3d.components.geometry.utils import merging_geometries_on_plane
from tidy3d.components.medium import PEC, LossyMetalMedium
from tidy3d.components.structure import Structure
from tidy3d.components.types import ArrayFloat1D, ArrayFloat2D, Axis, Shapely
from tidy3d.constants import inf

CORNER_ANGLE_THRESOLD = 0.1 * np.pi


class CornerFinderSpec(Tidy3dBaseModel):
    """Specification for corner detection on a 2D plane."""

    medium: Literal["metal", "dielectric", "all"] = pd.Field(
        "metal",
        title="Material Type For Corner Identification",
        description="Find corners of structures made of ``medium``, "
        "which can take value ``metal`` for PEC and lossy metal, ``dielectric`` "
        "for non-metallic materials, and ``all`` for all materials.",
    )

    angle_threshold: float = pd.Field(
        CORNER_ANGLE_THRESOLD,
        title="Angle Threshold In Corner Identification",
        description="A vertex is qualified as a corner if the angle spanned by its two edges "
        "is larger than the supplementary angle of "
        "this threshold value.",
        ge=0,
        lt=np.pi,
    )

    distance_threshold: Optional[pd.PositiveFloat] = pd.Field(
        None,
        title="Distance Threshold In Corner Identification",
        description="If not ``None`` and the distance of the vertex to its neighboring vertices "
        "is below the threshold value based on Douglas-Peucker algorithm, the vertex is disqualified as a corner.",
    )

    concave_resolution: Optional[pd.PositiveInt] = pd.Field(
        None,
        title="Concave Region Resolution.",
        description="Specifies number of steps to use for determining `dl_min` based on concave featues."
        "If set to ``None``, then the corresponding `dl_min` reduction is not applied.",
    )

    convex_resolution: Optional[pd.PositiveInt] = pd.Field(
        None,
        title="Convex Region Resolution.",
        description="Specifies number of steps to use for determining `dl_min` based on convex featues."
        "If set to ``None``, then the corresponding `dl_min` reduction is not applied.",
    )

    mixed_resolution: Optional[pd.PositiveInt] = pd.Field(
        None,
        title="Mixed Region Resolution.",
        description="Specifies number of steps to use for determining `dl_min` based on mixed featues."
        "If set to ``None``, then the corresponding `dl_min` reduction is not applied.",
    )

    @cached_property
    def _no_min_dl_override(self):
        return all(
            (
                self.concave_resolution is None,
                self.convex_resolution is None,
                self.mixed_resolution is None,
            )
        )

    @classmethod
    def _merged_pec_on_plane(
        cls,
        normal_axis: Axis,
        coord: float,
        structure_list: list[Structure],
        center: tuple[float, float] = [0, 0, 0],
        size: tuple[float, float, float] = [inf, inf, inf],
    ) -> list[tuple[Any, Shapely]]:
        """On a 2D plane specified by axis = `normal_axis` and coordinate `coord`, merge geometries made of PEC.

        Parameters
        ----------
        normal_axis : Axis
            Axis normal to the 2D plane.
        coord : float
            Position of plane along the normal axis.
        structure_list : List[Structure]
            List of structures present in simulation.
        center : Tuple[float, float] = [0, 0, 0]
            Center of the 2D plane (coordinate along ``axis`` is ignored)
        size : Tuple[float, float, float] = [inf, inf, inf]
            Size of the 2D plane (size along ``axis`` is ignored)

        Returns
        -------
        List[Tuple[Any, Shapely]]
            List of shapes and their property value on the plane after merging.
        """

        # Construct plane
        slice_center = list(center)
        slice_size = list(size)
        slice_center[normal_axis] = coord
        slice_size[normal_axis] = 0
        plane = Box(center=slice_center, size=slice_size)

        # prepare geometry and medium list
        geometry_list = [structure.geometry for structure in structure_list]
        # For metal, we don't distinguish between LossyMetal and PEC,
        # so they'll be merged to PEC. Other materials are considered as dielectric.
        medium_list = (structure.medium for structure in structure_list)
        medium_list = [
            PEC if (mat.is_pec or isinstance(mat, LossyMetalMedium)) else mat for mat in medium_list
        ]
        # merge geometries
        merged_geos = merging_geometries_on_plane(geometry_list, plane, medium_list)

        return merged_geos

    def _corners_and_convexity(
        self,
        normal_axis: Axis,
        coord: float,
        structure_list: list[Structure],
        ravel: bool,
    ) -> tuple[ArrayFloat2D, ArrayFloat1D]:
        """On a 2D plane specified by axis = `normal_axis` and coordinate `coord`, find out corners of merged
        geometries made of PEC.


        Parameters
        ----------
        normal_axis : Axis
            Axis normal to the 2D plane.
        coord : float
            Position of plane along the normal axis.
        structure_list : List[Structure]
            List of structures present in simulation.
        ravel : bool
            Whether to put the resulting corners in a single list or per polygon.

        Returns
        -------
        Tuple[ArrayFloat2D, ArrayFloat1D]
            Corner coordinates and their convexity.
        """

        # merge geometries
        merged_geos = self._merged_pec_on_plane(
            normal_axis=normal_axis, coord=coord, structure_list=structure_list
        )

        # corner finder
        corner_list = []
        convexity_list = []
        for mat, shapes in merged_geos:
            if self.medium != "all" and mat.is_pec != (self.medium == "metal"):
                continue
            polygon_list = ClipOperation.to_polygon_list(shapes)
            for poly in polygon_list:
                poly = poly.normalize().buffer(0)
                if self.distance_threshold is not None:
                    poly = poly.simplify(self.distance_threshold, preserve_topology=True)
                corners_xy, corners_convexity = self._filter_collinear_vertices(
                    list(poly.exterior.coords)
                )
                corner_list.append(corners_xy)
                convexity_list.append(corners_convexity)
                # in case the polygon has holes
                for poly_inner in poly.interiors:
                    corners_xy, corners_convexity = self._filter_collinear_vertices(
                        list(poly_inner.coords)
                    )
                    corner_list.append(corners_xy)
                    convexity_list.append(corners_convexity)

        if ravel and len(corner_list) > 0:
            corner_list = np.concatenate(corner_list)
            convexity_list = np.concatenate(convexity_list)

        return corner_list, convexity_list

    def corners(
        self,
        normal_axis: Axis,
        coord: float,
        structure_list: list[Structure],
    ) -> ArrayFloat2D:
        """On a 2D plane specified by axis = `normal_axis` and coordinate `coord`, find out corners of merged
        geometries made of `medium`.


        Parameters
        ----------
        normal_axis : Axis
            Axis normal to the 2D plane.
        coord : float
            Position of plane along the normal axis.
        structure_list : List[Structure]
            List of structures present in simulation.

        Returns
        -------
        ArrayFloat2D
            Corner coordinates.
        """

        corner_list, _ = self._corners_and_convexity(
            normal_axis=normal_axis, coord=coord, structure_list=structure_list, ravel=True
        )
        return corner_list

    def _filter_collinear_vertices(
        self, vertices: ArrayFloat2D
    ) -> tuple[ArrayFloat2D, ArrayFloat1D]:
        """Filter collinear vertices of a polygon, and return corners locations and their convexity.

        Parameters
        ----------
        vertices : ArrayFloat2D
            Polygon vertices from shapely.Polygon. The last vertex is identical to the 1st
            vertex to make a valid polygon.

        Returns
        -------
        ArrayFloat2D
            Corner coordinates.
        ArrayFloat1D
            Convexity of corners: True for outer corners, False for inner corners.
        """

        def normalize(v):
            return v / np.linalg.norm(v, axis=-1)[:, np.newaxis]

        # drop the last vertex, which is identical to the 1st one.
        vs_orig = np.array(vertices[:-1])
        # compute unit vector to next and previous vertex
        vs_next = np.roll(vs_orig, axis=0, shift=-1)
        vs_previous = np.roll(vs_orig, axis=0, shift=+1)
        unit_next = normalize(vs_next - vs_orig)
        unit_previous = normalize(vs_previous - vs_orig)
        # angle
        inner_product = np.sum(unit_next * unit_previous, axis=-1)
        inner_product = np.where(inner_product > 1, 1, inner_product)
        inner_product = np.where(inner_product < -1, -1, inner_product)
        angle = np.arccos(inner_product)
        num_vs = len(vs_orig)
        cross_product = np.cross(
            np.hstack([unit_next, np.zeros((num_vs, 1))]),
            np.hstack([unit_previous, np.zeros((num_vs, 1))]),
            axis=-1,
        )
        convexity = cross_product[:, 2] < 0
        ind_filter = angle <= np.pi - self.angle_threshold
        return vs_orig[ind_filter], convexity[ind_filter]



"""Defines the FDTD grid."""

from __future__ import annotations

from typing import Literal, Union

import numpy as np
import pydantic.v1 as pd

from tidy3d.components.base import Tidy3dBaseModel, cached_property
from tidy3d.components.data.data_array import DataArray, ScalarFieldDataArray, SpatialDataArray
from tidy3d.components.data.utils import UnstructuredGridDataset, UnstructuredGridDatasetType
from tidy3d.components.geometry.base import Box
from tidy3d.components.types import ArrayFloat1D, Axis, Coordinate, InterpMethod
from tidy3d.exceptions import SetupError

# data type of one dimensional coordinate array.
Coords1D = ArrayFloat1D


class Coords(Tidy3dBaseModel):
    """Holds data about a set of x,y,z positions on a grid.

    Example
    -------
    >>> x = np.linspace(-1, 1, 10)
    >>> y = np.linspace(-1, 1, 11)
    >>> z = np.linspace(-1, 1, 12)
    >>> coords = Coords(x=x, y=y, z=z)
    """

    x: Coords1D = pd.Field(
        ..., title="X Coordinates", description="1-dimensional array of x coordinates."
    )

    y: Coords1D = pd.Field(
        ..., title="Y Coordinates", description="1-dimensional array of y coordinates."
    )

    z: Coords1D = pd.Field(
        ..., title="Z Coordinates", description="1-dimensional array of z coordinates."
    )

    @property
    def to_dict(self):
        """Return a dict of the three Coord1D objects as numpy arrays."""
        return {key: self.dict()[key] for key in "xyz"}

    @property
    def to_list(self):
        """Return a list of the three Coord1D objects as numpy arrays."""
        return list(self.to_dict.values())

    @cached_property
    def cell_sizes(self) -> SpatialDataArray:
        """Returns the sizes of the cells in each coordinate array as a dictionary."""
        cell_sizes = {}

        coord_dict = self.to_dict
        for dim in "xyz":
            if len(coord_dict[dim]) > 1:
                diff = coord_dict[dim][1:] - coord_dict[dim][0:-1]

                diff_left = np.pad(diff, ((1, 0)), mode="edge")
                diff_right = np.pad(diff, ((0, 1)), mode="edge")

                diff_avg = 0.5 * (diff_left + diff_right)
                cell_sizes[dim] = diff_avg
            else:
                cell_sizes[dim] = 1

        return cell_sizes

    @cached_property
    def cell_size_meshgrid(self):
        """Returns an N-dimensional grid where N is the number of coordinate arrays that have more than one
        element. Each grid element corresponds to the size of the mesh cell in N-dimensions and 1 for N=0."""
        coord_dict = self.to_dict

        cell_size_meshgrid = np.squeeze(np.ones(tuple(len(coord_dict[dim]) for dim in "xyz")))
        meshgrid_elements = [
            size for dim, size in self.cell_sizes.items() if len(coord_dict[dim]) > 1
        ]

        if len(meshgrid_elements) > 1:
            meshgrid = np.meshgrid(*meshgrid_elements, indexing="ij")
            for idx in range(len(meshgrid)):
                cell_size_meshgrid *= np.reshape(meshgrid[idx], cell_size_meshgrid.shape)
        elif len(meshgrid_elements) == 1:
            cell_size_meshgrid = meshgrid_elements[0]

        return cell_size_meshgrid

    def _interp_from_xarray(
        self,
        array: Union[SpatialDataArray, ScalarFieldDataArray],
        interp_method: InterpMethod,
        fill_value: Union[Literal["extrapolate"], float] = "extrapolate",
    ) -> Union[SpatialDataArray, ScalarFieldDataArray]:
        """
        Similar to ``xarrray.DataArray.interp`` with 2 enhancements:

            1) Check if the coordinate of the supplied data are in monotonically increasing order.
            If they are, apply the faster ``assume_sorted=True``.

            2) For axes of single entry, instead of error, apply ``isel()`` along the axis.

        Parameters
        ----------
        array : Union[:class:`.SpatialDataArray`, :class:`.ScalarFieldDataArray`]
            Supplied scalar dataset
        interp_method : :class:`.InterpMethod`
            Interpolation method.
        fill_value : Union[Literal['extrapolate'], float] = "extrapolate"
            Value used to fill in for points outside the data range. If set to 'extrapolate',
            values will be extrapolated into those regions using the "nearest" method.

        Returns
        -------
        Union[:class:`.SpatialDataArray`, :class:`.ScalarFieldDataArray`]
            The interpolated spatial dataset.

        Note
        ----
        This method is called from a :class:`Coords` instance with the array to be interpolated as
        an argument, not the other way around.
        """
        # Check which axes need interpolation or selection
        interp_ax = []
        isel_ax = []
        for ax in "xyz":
            if array.sizes[ax] == 1:
                isel_ax.append(ax)
            else:
                interp_ax.append(ax)

        # apply iselection for the axis containing single entry
        if len(isel_ax) > 0:
            array = array.isel({ax: [0] * len(self.to_dict[ax]) for ax in isel_ax})
            array = array.assign_coords({ax: self.to_dict[ax] for ax in isel_ax})
            if len(interp_ax) == 0:
                return array

        # Apply interp for the rest
        is_sorted = all(np.all(np.diff(array.coords[f]) > 0) for f in interp_ax)
        interp_param = {
            "method": interp_method,
            "assume_sorted": is_sorted,
            "kwargs": {
                "bounds_error": False,
                "fill_value": fill_value,
            },
        }

        # Mark extrapolated points with nan's to fill in later
        if fill_value == "extrapolate" and interp_method != "nearest":
            interp_param["kwargs"]["fill_value"] = np.nan

        # interpolation
        interp_array = array.interp({ax: self.to_dict[ax] for ax in interp_ax}, **interp_param)

        # Fill in nan's with nearest values
        if fill_value == "extrapolate" and interp_method != "nearest":
            interp_param["method"] = "nearest"
            interp_param["kwargs"]["fill_value"] = "extrapolate"
            nearest_array = array.interp({ax: self.to_dict[ax] for ax in interp_ax}, **interp_param)
            interp_array.values[:] = np.where(
                np.isnan(interp_array.values), nearest_array.values, interp_array.values
            )

        return interp_array

    def _interp_from_unstructured(
        self,
        array: UnstructuredGridDatasetType,
        interp_method: InterpMethod,
        fill_value: Union[Literal["extrapolate"], float] = "extrapolate",
    ) -> SpatialDataArray:
        """
        Interpolate from untructured grid onto a Cartesian one.

        Parameters
        ----------
        array : Union[class:`.TriangularGridDataset`, class:`.TetrahedralGridDataset`]
            Supplied scalar dataset
        interp_method : :class:`.InterpMethod`
            Interpolation method.
        fill_value : Union[Literal['extrapolate'], float] = "extrapolate"
            Value used to fill in for points outside the data range. If set to 'extrapolate',
            values will be extrapolated into those regions using the "nearest" method.

        Returns
        -------
        :class:`.SpatialDataArray`
            The interpolated spatial dataset.

        Note
        ----
        This method is called from a :class:`Coords` instance with the array to be interpolated as
        an argument, not the other way around.
        """
        interp_array = array.interp(
            **{ax: self.to_dict[ax] for ax in "xyz"}, method=interp_method, fill_value=fill_value
        )

        return interp_array

    def spatial_interp(
        self,
        array: Union[SpatialDataArray, ScalarFieldDataArray, UnstructuredGridDatasetType],
        interp_method: InterpMethod,
        fill_value: Union[Literal["extrapolate"], float] = "extrapolate",
    ) -> Union[SpatialDataArray, ScalarFieldDataArray]:
        """
        Similar to ``xarrray.DataArray.interp`` with 2 enhancements:

            1) (if input data is an ``xarrray.DataArray``) Check if the coordinate of the supplied
            data are in monotonically increasing order. If they are, apply the faster
            ``assume_sorted=True``.

            2) Data is assumed invariant along zero-size dimensions (if any).

        Parameters
        ----------
        array : Union[
                :class:`.SpatialDataArray`,
                :class:`.ScalarFieldDataArray`,
                :class:`.TriangularGridDataset`,
                :class:`.TetrahedralGridDataset`,
        ]
            Supplied scalar dataset
        interp_method : :class:`.InterpMethod`
            Interpolation method.
        fill_value : Union[Literal['extrapolate'], float] = "extrapolate"
            Value used to fill in for points outside the data range. If set to 'extrapolate',
            values will be extrapolated into those regions using the "nearest" method.

        Returns
        -------
        Union[:class:`.SpatialDataArray`, :class:`.ScalarFieldDataArray`]
            The interpolated spatial dataset.

        Note
        ----
        This method is called from a :class:`Coords` instance with the array to be interpolated as
        an argument, not the other way around.
        """

        # Check for empty dimensions
        result_coords = dict(self.to_dict)
        if any(len(v) == 0 for v in result_coords.values()):
            if isinstance(array, (SpatialDataArray, ScalarFieldDataArray)):
                for c in array.coords:
                    if c not in result_coords:
                        result_coords[c] = array.coords[c].values
            result_shape = tuple(len(v) for v in result_coords.values())
            result = DataArray(np.empty(result_shape, dtype=array.dtype), coords=result_coords)
            return result

        # interpolation
        if isinstance(array, UnstructuredGridDataset):
            return self._interp_from_unstructured(
                array=array, interp_method=interp_method, fill_value=fill_value
            )
        return self._interp_from_xarray(
            array=array, interp_method=interp_method, fill_value=fill_value
        )


class FieldGrid(Tidy3dBaseModel):
    """Holds the grid data for a single field.

    Example
    -------
    >>> x = np.linspace(-1, 1, 10)
    >>> y = np.linspace(-1, 1, 11)
    >>> z = np.linspace(-1, 1, 12)
    >>> coords = Coords(x=x, y=y, z=z)
    >>> field_grid = FieldGrid(x=coords, y=coords, z=coords)
    """

    x: Coords = pd.Field(
        ...,
        title="X Positions",
        description="x,y,z coordinates of the locations of the x-component of a vector field.",
    )

    y: Coords = pd.Field(
        ...,
        title="Y Positions",
        description="x,y,z coordinates of the locations of the y-component of a vector field.",
    )

    z: Coords = pd.Field(
        ...,
        title="Z Positions",
        description="x,y,z coordinates of the locations of the z-component of a vector field.",
    )


class YeeGrid(Tidy3dBaseModel):
    """Holds the yee grid coordinates for each of the E and H positions.

    Example
    -------
    >>> x = np.linspace(-1, 1, 10)
    >>> y = np.linspace(-1, 1, 11)
    >>> z = np.linspace(-1, 1, 12)
    >>> coords = Coords(x=x, y=y, z=z)
    >>> field_grid = FieldGrid(x=coords, y=coords, z=coords)
    >>> yee_grid = YeeGrid(E=field_grid, H=field_grid)
    >>> Ex_coords = yee_grid.E.x
    """

    E: FieldGrid = pd.Field(
        ...,
        title="Electric Field Grid",
        description="Coordinates of the locations of all three components of the electric field.",
    )

    H: FieldGrid = pd.Field(
        ...,
        title="Electric Field Grid",
        description="Coordinates of the locations of all three components of the magnetic field.",
    )

    @property
    def grid_dict(self):
        """The Yee grid coordinates associated to various field components as a dictionary."""
        return {
            "Ex": self.E.x,
            "Ey": self.E.y,
            "Ez": self.E.z,
            "Hx": self.H.x,
            "Hy": self.H.y,
            "Hz": self.H.z,
        }


class Grid(Tidy3dBaseModel):
    """Contains all information about the spatial positions of the FDTD grid.

    Example
    -------
    >>> x = np.linspace(-1, 1, 10)
    >>> y = np.linspace(-1, 1, 11)
    >>> z = np.linspace(-1, 1, 12)
    >>> coords = Coords(x=x, y=y, z=z)
    >>> grid = Grid(boundaries=coords)
    >>> centers = grid.centers
    >>> sizes = grid.sizes
    >>> yee_grid = grid.yee
    """

    boundaries: Coords = pd.Field(
        ...,
        title="Boundary Coordinates",
        description="x,y,z coordinates of the boundaries between cells, defining the FDTD grid.",
    )

    @staticmethod
    def _avg(coords1d: Coords1D):
        """Return average positions of an array of 1D coordinates."""
        return (coords1d[1:] + coords1d[:-1]) / 2.0

    @staticmethod
    def _min(coords1d: Coords1D):
        """Return minus positions of 1D coordinates."""
        return coords1d[:-1]

    @property
    def centers(self) -> Coords:
        """Return centers of the cells in the :class:`Grid`.

        Returns
        -------
        :class:`Coords`
            centers of the FDTD cells in x,y,z stored as :class:`Coords` object.

        Example
        -------
        >>> x = np.linspace(-1, 1, 10)
        >>> y = np.linspace(-1, 1, 11)
        >>> z = np.linspace(-1, 1, 12)
        >>> coords = Coords(x=x, y=y, z=z)
        >>> grid = Grid(boundaries=coords)
        >>> centers = grid.centers
        """
        return Coords(**{key: self._avg(val) for key, val in self.boundaries.to_dict.items()})

    @property
    def sizes(self) -> Coords:
        """Return sizes of the cells in the :class:`Grid`.

        Returns
        -------
        :class:`Coords`
            Sizes of the FDTD cells in x,y,z stored as :class:`Coords` object.

        Example
        -------
        >>> x = np.linspace(-1, 1, 10)
        >>> y = np.linspace(-1, 1, 11)
        >>> z = np.linspace(-1, 1, 12)
        >>> coords = Coords(x=x, y=y, z=z)
        >>> grid = Grid(boundaries=coords)
        >>> sizes = grid.sizes
        """
        return Coords(**{key: np.diff(val) for key, val in self.boundaries.to_dict.items()})

    @property
    def num_cells(self) -> tuple[int, int, int]:
        """Return sizes of the cells in the :class:`Grid`.

        Returns
        -------
        tuple[int, int, int]
            Number of cells in the grid in the x, y, z direction.

        Example
        -------
        >>> x = np.linspace(-1, 1, 10)
        >>> y = np.linspace(-1, 1, 11)
        >>> z = np.linspace(-1, 1, 12)
        >>> coords = Coords(x=x, y=y, z=z)
        >>> grid = Grid(boundaries=coords)
        >>> Nx, Ny, Nz = grid.num_cells
        """
        return [len(self.boundaries.dict()[dim]) - 1 for dim in "xyz"]

    @property
    def min_size(self) -> float:
        """Return minimal cells size in all dimensions.

        Returns
        -------
        float
            Minimal cells size in all dimensions.
        """
        return float(min(min(sizes) for sizes in self.sizes.to_list))

    @property
    def max_size(self) -> float:
        """Return maximal cells size in all dimensions.

        Returns
        -------
        float
            Maximal cells size in all dimensions.
        """
        return float(max(max(sizes) for sizes in self.sizes.to_list))

    @property
    def info(self) -> dict:
        """Dictionary collecting various properties of the grids."""
        num_cells = self.num_cells
        total_cells = int(np.prod(num_cells))
        return {
            "Nx": num_cells[0],
            "Ny": num_cells[1],
            "Nz": num_cells[2],
            "grid_points": total_cells,
            "min_grid_size": self.min_size,
            "max_grid_size": self.max_size,
            "computational_complexity": total_cells / self.min_size,
        }

    @property
    def _primal_steps(self) -> Coords:
        """Return primal steps of the cells in the :class:`Grid`.

        Returns
        -------
        :class:`Coords`
            Distances between each of the cell boundaries along each dimension.
        """
        return self.sizes

    @property
    def _dual_steps(self) -> Coords:
        """Return dual steps of the cells in the :class:`Grid`.

        Returns
        -------
        :class:`Coords`
            Distances between each of the cell centers along each dimension, with periodicity
            applied.
        """

        primal_steps = {dim: self._primal_steps.dict()[dim] for dim in "xyz"}
        dsteps = {key: (psteps + np.roll(psteps, 1)) / 2 for (key, psteps) in primal_steps.items()}

        return Coords(**dsteps)

    @property
    def yee(self) -> YeeGrid:
        """Return the :class:`YeeGrid` defining the yee cell locations for this :class:`Grid`.


        Returns
        -------
        :class:`YeeGrid`
            Stores coordinates of all of the components on the yee lattice.

        Example
        -------
        >>> x = np.linspace(-1, 1, 10)
        >>> y = np.linspace(-1, 1, 11)
        >>> z = np.linspace(-1, 1, 12)
        >>> coords = Coords(x=x, y=y, z=z)
        >>> grid = Grid(boundaries=coords)
        >>> yee_cells = grid.yee
        >>> Ex_positions = yee_cells.E.x
        """
        yee_e_kwargs = {key: self._yee_e(axis=axis) for axis, key in enumerate("xyz")}
        yee_h_kwargs = {key: self._yee_h(axis=axis) for axis, key in enumerate("xyz")}

        yee_e = FieldGrid(**yee_e_kwargs)
        yee_h = FieldGrid(**yee_h_kwargs)
        return YeeGrid(E=yee_e, H=yee_h)

    def __getitem__(self, coord_key: str) -> Coords:
        """quickly get the grid element by grid[key]."""

        coord_dict = {
            "centers": self.centers,
            "sizes": self.sizes,
            "boundaries": self.boundaries,
            "Ex": self.yee.E.x,
            "Ey": self.yee.E.y,
            "Ez": self.yee.E.z,
            "Hx": self.yee.H.x,
            "Hy": self.yee.H.y,
            "Hz": self.yee.H.z,
        }
        if coord_key not in coord_dict:
            raise SetupError(f"key {coord_key} not found in grid with {list(coord_dict.keys())} ")

        return coord_dict.get(coord_key)

    def _yee_e(self, axis: Axis):
        """E field yee lattice sites for axis."""

        boundary_coords = self.boundaries.to_dict

        # initially set all to the minus bounds
        yee_coords = {key: self._min(val) for key, val in boundary_coords.items()}

        # average the axis index between the cell boundaries
        key = "xyz"[axis]
        yee_coords[key] = self._avg(boundary_coords[key])

        return Coords(**yee_coords)

    def _yee_h(self, axis: Axis):
        """H field yee lattice sites for axis."""

        boundary_coords = self.boundaries.to_dict

        # initially set all to centers
        yee_coords = {key: self._avg(val) for key, val in boundary_coords.items()}

        # set the axis index to the minus bounds
        key = "xyz"[axis]
        yee_coords[key] = self._min(boundary_coords[key])

        return Coords(**yee_coords)

    def discretize_inds(self, box: Box, extend: bool = False) -> list[tuple[int, int]]:
        """Start and stopping indexes for the cells that intersect with a :class:`Box`.

        Parameters
        ----------
        box : :class:`Box`
            Rectangular geometry within simulation to discretize.
        extend : bool = False
            If ``True``, ensure that the returned indexes extend sufficiently in every direction to
            be able to interpolate any field component at any point within the ``box``, for field
            components sampled on the Yee grid.

        Returns
        -------
        List[Tuple[int, int]]
            The (start, stop) indexes of the cells that intersect with ``box`` in each of the three
            dimensions.
        """

        pts_min, pts_max = box.bounds
        boundaries = self.boundaries

        inds_list = []

        # for each dimension
        for axis, (pt_min, pt_max) in enumerate(zip(pts_min, pts_max)):
            bound_coords = np.array(boundaries.to_list[axis])
            if pt_min > pt_max:
                raise AssertionError("min point was greater than max point")

            # index of smallest coord greater than pt_max
            inds_gt_pt_max = np.where(bound_coords > pt_max)[0]
            ind_max = len(bound_coords) - 1 if len(inds_gt_pt_max) == 0 else inds_gt_pt_max[0]

            # index of largest coord less than or equal to pt_min
            inds_leq_pt_min = np.where(bound_coords <= pt_min)[0]
            ind_min = 0 if len(inds_leq_pt_min) == 0 else inds_leq_pt_min[-1]

            # handle extensions
            if ind_max > ind_min and extend:
                # Left side
                if box.bounds[0][axis] < self.centers.to_list[axis][ind_min]:
                    # Box bounds on the left side are to the left of the closest grid center
                    ind_min -= 1

                # We always need an extra pixel on the right for the tangential components
                ind_max += 1

            # store indexes
            inds_list.append((ind_min, ind_max))

        return inds_list

    def extended_subspace(
        self,
        axis: Axis,
        ind_beg: int = 0,
        ind_end: int = 0,
        periodic: bool = True,
    ) -> Coords1D:
        """Pick a subspace of 1D boundaries within ``range(ind_beg, ind_end)``. If any indexes lie
        outside of the grid boundaries array, padding is used based on the boundary conditions.
        For periodic BCs, the zeroth and last element of the grid boundaries are identified.
        For other BCs, the zeroth and last element of the boundaries are a reflection plane.

        Parameters
        ----------
        axis : Axis
            Axis index along which to pick the subspace.
        ind_beg : int = 0
            Starting index for the subspace.
        ind_end : int = 0
            Ending index for the subspace.
        periodic : bool = True
            Whether to pad out of bounds indexes with a periodic or reflected pattern.

        Returns
        -------
        Coords1D
            The subspace of the grid along ``axis``.
        """

        coords = self.boundaries.to_list[axis]
        padded_coords = coords
        num_cells = coords.size - 1

        reverse = True
        while ind_beg < 0:
            if periodic or not reverse:
                offset = padded_coords[0] - coords[-1]
                padded_coords = np.concatenate([coords[:-1] + offset, padded_coords])
                reverse = True
            else:
                offset = padded_coords[0] + coords[0]
                padded_coords = np.concatenate([offset - coords[:0:-1], padded_coords])
                reverse = False
            ind_beg += num_cells
            ind_end += num_cells

        reverse = True
        while ind_end >= padded_coords.size:
            if periodic or not reverse:
                offset = padded_coords[-1] - coords[0]
                padded_coords = np.concatenate([padded_coords, coords[1:] + offset])
                reverse = True
            else:
                offset = padded_coords[-1] + coords[-1]
                padded_coords = np.concatenate([padded_coords, offset - coords[-2::-1]])
                reverse = False

        return padded_coords[ind_beg:ind_end]

    def snap_to_box_zero_dim(self, box: Box):
        """Snap a grid to an exact box position for dimensions for which the box is size 0.
        If the box location is outside of the grid, an error is raised.

        Parameters
        ----------
        box : :class:`Box`
            Box to use for the zero dim check.

        Returns
        -------
        class:`Grid`
            Snapped copy of the grid.
        """

        boundary_dict = self.boundaries.to_dict.copy()
        for dim, center, size in zip("xyz", box.center, box.size):
            # Overwrite grid boundaries with box center if box is size 0 along dimension
            if size == 0:
                if boundary_dict[dim][0] > center or boundary_dict[dim][-1] < center:
                    raise ValueError("Cannot snap grid to box center outside of grid domain.")
                boundary_dict[dim] = np.array([center, center])
        return self.updated_copy(boundaries=Coords(**boundary_dict))

    def _translated_copy(self, vector: Coordinate) -> Grid:
        """Translate the grid by a vector. Not officially supported as resulting
        grid may not be aligned with original Yee grid."""
        boundaries = Coords(
            x=self.boundaries.x + vector[0],
            y=self.boundaries.y + vector[1],
            z=self.boundaries.z + vector[2],
        )
        return self.updated_copy(boundaries=boundaries)



"""Defines mode solver simulation class."""

from __future__ import annotations

from typing import Optional, Union

import numpy as np
import pydantic.v1 as pd

from tidy3d.components.base import cached_property
from tidy3d.components.boundary import BoundarySpec
from tidy3d.components.geometry.base import Box
from tidy3d.components.grid.grid import Grid
from tidy3d.components.grid.grid_spec import GridSpec
from tidy3d.components.mode_spec import ModeSpec
from tidy3d.components.monitor import ModeMonitor, ModeSolverMonitor, PermittivityMonitor
from tidy3d.components.simulation import (
    AbstractYeeGridSimulation,
    Simulation,
    validate_boundaries_for_zero_dims,
)
from tidy3d.components.source.field import ModeSource
from tidy3d.components.types import TYPE_TAG_STR, Ax, Direction, EMField, FreqArray
from tidy3d.constants import C_0
from tidy3d.exceptions import SetupError, ValidationError
from tidy3d.log import log

from .mode_solver import ModeSolver

ModeSimulationMonitorType = PermittivityMonitor

# dummy run time for conversion to FDTD sim
# should be very small -- otherwise, generating tmesh will fail or take a long time
RUN_TIME = 1e-30

MODE_PLANE_TYPE = Union[Box, ModeSource, ModeMonitor, ModeSolverMonitor]


# attributes shared between ModeSimulation class and ModeSolver class
MODE_SIM_MODE_SOLVER_SHARED_ATTRS = [
    "plane",
    "mode_spec",
    "freqs",
    "direction",
    "colocate",
    "fields",
]
# attributes shared between ModeSimulation class and AbstractYeeGridSimulation
# excludes things like monitors and sources which are not present in a ModeSimulation
MODE_SIM_YEE_SIM_SHARED_ATTRS = [
    "center",
    "size",
    "medium",
    "structures",
    "symmetry",
    "boundary_spec",
    "version",
    "plot_length_units",
    "lumped_elements",
    "subpixel",
    "simulation_type",
    "post_norm",
]


class ModeSimulation(AbstractYeeGridSimulation):
    """
    Simulation class for solving electromagnetic eigenmodes in a 2D plane with
    translational invariance in the third dimension. If the field ``plane`` is
    specified, the domain for mode solving is the intersection between the ``plane``
    and the simulation geometry. If the simulation geometry is 2D (has zero size
    in one dimension) and the ``plane`` is ``None``, then the domain for mode solving
    is the entire 2D geometry.

    The ``symmetry`` field can be used to enforce reflection symmetry across planes
    through the ``center`` of the simulation. Each component of the ``symmetry`` field
    is only used if the ``center`` of the ``plane`` and the simulation geometry
    coincide in that component. Symmetry normal to the mode solving domain has no
    effect; the field ``filter_pol`` in :class:`.ModeSpec` can be used here instead.

    Example
    -------
    >>> from tidy3d import C_0, ModeSpec, BoundarySpec, Boundary
    >>> lambda0 = 1550e-9
    >>> freq0 = C_0 / lambda0
    >>> freqs = [freq0]
    >>> sim_size = lambda0, lambda0, 0
    >>> mode_spec = ModeSpec(num_modes=4)
    >>> boundary_spec = BoundarySpec(
    ...     x=Boundary.pec(),
    ...     y=Boundary.pec(),
    ...     z=Boundary.periodic()
    ... )
    >>> sim = ModeSimulation(
    ...     size=sim_size,
    ...     freqs=freqs,
    ...     mode_spec=mode_spec,
    ...     boundary_spec=boundary_spec
    ... )

    See Also
    --------

    :class:`ModeSource`:
        Injects current source to excite modal profile on finite extent plane.

    **Notebooks:**
        * `Waveguide Y junction <../../notebooks/YJunction.html>`_
        * `Photonic crystal waveguide polarization filter <../../../notebooks/PhotonicCrystalWaveguidePolarizationFilter.html>`_

    **Lectures:**
        * `Prelude to Integrated Photonics Simulation: Mode Injection <https://www.flexcompute.com/fdtd101/Lecture-4-Prelude-to-Integrated-Photonics-Simulation-Mode-Injection/>`_
    """

    mode_spec: ModeSpec = pd.Field(
        ...,
        title="Mode specification",
        description="Container with specifications about the modes to be solved for.",
    )

    freqs: FreqArray = pd.Field(
        ..., title="Frequencies", description="A list of frequencies at which to solve."
    )

    direction: Direction = pd.Field(
        "+",
        title="Propagation direction",
        description="Direction of waveguide mode propagation along the axis defined by its normal "
        "dimension.",
    )

    colocate: bool = pd.Field(
        True,
        title="Colocate fields",
        description="Toggle whether fields should be colocated to grid cell boundaries (i.e. "
        "primal grid nodes). Default is ``True``.",
    )

    fields: tuple[EMField, ...] = pd.Field(
        ["Ex", "Ey", "Ez", "Hx", "Hy", "Hz"],
        title="Field Components",
        description="Collection of field components to store in the monitor. Note that some "
        "methods like ``flux``, ``dot`` require all tangential field components, while others "
        "like ``mode_area`` require all E-field components.",
    )

    boundary_spec: BoundarySpec = pd.Field(
        BoundarySpec(),
        title="Boundaries",
        description="Specification of boundary conditions along each dimension. If ``None``, "
        "PML boundary conditions are applied on all sides. This behavior is for "
        "consistency with FDTD simulations; however, please note that the mode solver "
        "terminates the mode plane with PEC boundary. The 'ModeSpec' can be used to "
        "apply PML layers in the mode solver.",
    )

    monitors: tuple[ModeSimulationMonitorType, ...] = pd.Field(
        (),
        title="Monitors",
        description="Tuple of monitors in the simulation. "
        "Note: monitor names are used to access data after simulation is run.",
    )

    sources: tuple[()] = pd.Field(
        (),
        title="Sources",
        description="Sources in the simulation. Note: sources are not supported in mode "
        "simulations.",
    )

    grid_spec: GridSpec = pd.Field(
        GridSpec(),
        title="Grid Specification",
        description="Specifications for the simulation grid along each of the three directions.",
    )

    plane: MODE_PLANE_TYPE = pd.Field(
        None,
        title="Plane",
        description="Cross-sectional plane in which the mode will be computed. "
        "If provided, the computational domain will be the intersection between "
        "the provided ``plane`` and the simulation geometry. "
        "If ``None``, the simulation must be 2D, and the plane will be the entire "
        "simulation geometry.",
        discriminator=TYPE_TAG_STR,
    )

    @pd.validator("plane", always=True)
    def is_plane(cls, val, values):
        """Raise validation error if not planar."""
        if val is None:
            sim_center = values.get("center")
            sim_size = values.get("size")
            val = Box(size=sim_size, center=sim_center)
            if val.size.count(0.0) != 1:
                raise ValidationError(
                    "If the 'ModeSimulation' geometry is not planar, "
                    "then 'plane' must be specified."
                )
            return val
        if val.size.count(0.0) != 1:
            raise ValidationError(f"'ModeSimulation.plane' must be planar, given 'size={val}'")
        return val

    @pd.validator("plane", always=True)
    def plane_in_sim_bounds(cls, val, values):
        """Check that the plane is at least partially inside the simulation bounds."""
        sim_center = values.get("center")
        sim_size = values.get("size")
        sim_box = Box(size=sim_size, center=sim_center)

        if not sim_box.intersects(val):
            raise SetupError("'ModeSimulation.plane' must intersect 'ModeSimulation.geometry.")
        return val

    def _post_init_validators(self) -> None:
        """Call validators taking `self` that get run after init."""
        _ = self._mode_solver
        _ = self.grid

    @pd.validator("grid_spec", always=True)
    def _validate_auto_grid_wavelength(cls, val, values):
        """Handle the case where grid_spec is auto and wavelength is not provided."""
        # this is handled instead post-init to ensure freqs is defined
        return val

    @cached_property
    def _mode_solver(self) -> ModeSolver:
        """Convert the :class:`.ModeSimulation` to a :class:`.ModeSolver`."""
        kwargs = {key: getattr(self, key) for key in MODE_SIM_MODE_SOLVER_SHARED_ATTRS}
        return ModeSolver(simulation=self._as_fdtd_sim, **kwargs)

    def run_local(self):
        """Run locally."""
        from .data.sim_data import ModeSimulationData

        # repeat the calculation every time, in case use_local_subpixel changed
        self._invalidate_solver_cache()

        modes_raw = self._mode_solver.data_raw
        return ModeSimulationData(simulation=self, modes_raw=modes_raw)

    @cached_property
    def grid(self) -> Grid:
        """Grid spatial locations and information as defined by `grid_spec`.

        Returns
        -------
        :class:`.Grid`
            :class:`.Grid` storing the spatial locations relevant to the simulation.
        """
        return self._as_fdtd_sim.grid

    @cached_property
    def _as_fdtd_sim(self) -> Simulation:
        """Convert :class:`.ModeSimulation` to :class:`.Simulation`.
        This should only be used to obtain the same material properties
        for mode solving or related purposes; the sources and monitors of the
        resulting simulation are not meaningful."""

        grid_spec = self.grid_spec
        if grid_spec.auto_grid_used and grid_spec.wavelength is None:
            min_wvl = C_0 / np.max(self.freqs)
            log.info(
                f"Auto meshing using wavelength {min_wvl:1.4f} defined from "
                "largest of 'ModeSimulation.freqs'."
            )
            grid_spec = grid_spec.updated_copy(wavelength=min_wvl)

        kwargs = {key: getattr(self, key) for key in MODE_SIM_YEE_SIM_SHARED_ATTRS}
        return Simulation(
            **kwargs,
            run_time=RUN_TIME,
            grid_spec=grid_spec,
            monitors=[],
        )

    @classmethod
    def from_simulation(
        cls,
        simulation: AbstractYeeGridSimulation,
        wavelength: Optional[pd.PositiveFloat] = None,
        **kwargs,
    ) -> ModeSimulation:
        """Creates :class:`.ModeSimulation` from a :class:`.AbstractYeeGridSimulation`.

        Parameters
        ----------
        simulation: :class:`.AbstractYeeGridSimulation`
            Starting simulation defining structures, grid, etc.
        wavelength: Optional[pd.PositiveFloat]
            Wavelength used for automatic grid generation. Required if auto grid
            is used in ``grid_spec``.
        **kwargs
            Other arguments passed to new mode simulation instance.

        Returns
        -------
        :class:`.ModeSimulation`
            Mode simulation reproducing the material and grid properties
            of the original simulation.
        """
        if isinstance(simulation, ModeSimulation):
            return simulation.updated_copy(**kwargs)

        grid_spec = simulation.grid_spec
        if grid_spec.auto_grid_used:
            if wavelength is None:
                if grid_spec.wavelength is None:
                    raise ValidationError(
                        "Automatic grid generation requires specifying "
                        "'wavelength' or 'grid_spec.wavelength'."
                    )
            else:
                if grid_spec.wavelength is not None and grid_spec.wavelength != wavelength:
                    log.warning("Replacing 'grid_spec.wavelength' with provided 'wavelength'.")
                grid_spec = grid_spec.updated_copy(wavelength=wavelength)

        sim_kwargs = {key: getattr(simulation, key) for key in MODE_SIM_YEE_SIM_SHARED_ATTRS}

        mode_sim = cls(
            grid_spec=grid_spec,
            **sim_kwargs,
            **kwargs,
        )

        return mode_sim

    @cached_property
    def reduced_simulation_copy(self) -> ModeSimulation:
        """Strip objects not used by the mode solver from simulation object.
        This might significantly reduce upload time in the presence of custom mediums.
        """
        reduced_mode_solver = self._mode_solver.reduced_simulation_copy
        return self.from_mode_solver(reduced_mode_solver)

    @classmethod
    def from_mode_solver(
        cls, mode_solver: ModeSolver, wavelength: Optional[pd.PositiveFloat] = None
    ) -> ModeSimulation:
        """Creates :class:`.ModeSimulation` from a :class:`.ModeSolver`.

        Parameters
        ----------
        simulation: :class:`.AbstractYeeGridSimulation`
            Starting simulation defining structures, grid, etc.
        wavelength: Optional[pd.PositiveFloat]
            Wavelength used for automatic grid generation. Required if auto grid
            is used in ``grid_spec``.

        Returns
        -------
        :class:`.ModeSimulation`
            Mode simulation reproducing the material and grid properties
            of the original simulation.
        """

        kwargs = {key: getattr(mode_solver, key) for key in MODE_SIM_MODE_SOLVER_SHARED_ATTRS}
        mode_sim = cls.from_simulation(
            simulation=mode_solver.simulation, wavelength=wavelength, **kwargs
        )
        return mode_sim

    def plot(
        self,
        x: Optional[float] = None,
        y: Optional[float] = None,
        z: Optional[float] = None,
        ax: Ax = None,
        source_alpha: Optional[float] = 0,
        monitor_alpha: Optional[float] = 0,
        lumped_element_alpha: Optional[float] = 0,
        hlim: Optional[tuple[float, float]] = None,
        vlim: Optional[tuple[float, float]] = None,
        fill_structures: bool = True,
        **patch_kwargs,
    ) -> Ax:
        """Plot the mode simulation. If any of ``x``, ``y``, or ``z`` is provided, the potentially
        larger FDTD simulation containing the mode plane is plotted at the desired location.
        Otherwise, the mode plane is plotted by default.

        Parameters
        ----------
        fill_structures : bool = True
            Whether to fill structures with color or just draw outlines.
        x : float = None
            position of plane in x direction, only one of x, y, z must be specified to define plane.
        y : float = None
            position of plane in y direction, only one of x, y, z must be specified to define plane.
        z : float = None
            position of plane in z direction, only one of x, y, z must be specified to define plane.
        source_alpha : float = 0
            Opacity of the sources. If ``None``, uses Tidy3d default.
        monitor_alpha : float = 0
            Opacity of the monitors. If ``None``, uses Tidy3d default.
        lumped_element_alpha : float = 0
            Opacity of the lumped elements. If ``None``, uses Tidy3d default.
        ax : matplotlib.axes._subplots.Axes = None
            Matplotlib axes to plot on, if not specified, one is created.
        hlim : Tuple[float, float] = None
            The x range if plotting on xy or xz planes, y range if plotting on yz plane.
        vlim : Tuple[float, float] = None
            The z range if plotting on xz or yz planes, y plane if plotting on xy plane.

        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.
        """

        if x is not None or y is not None or z is not None:
            return super().plot(
                x=x,
                y=y,
                z=z,
                ax=ax,
                source_alpha=source_alpha,
                monitor_alpha=monitor_alpha,
                lumped_element_alpha=lumped_element_alpha,
                hlim=hlim,
                vlim=vlim,
                fill_structures=fill_structures,
                **patch_kwargs,
            )
        return self._mode_solver.plot(
            ax=ax,
            hlim=hlim,
            vlim=vlim,
            fill_structures=fill_structures,
            **patch_kwargs,
        )

    def plot_mode_plane(
        self,
        ax: Ax = None,
        **patch_kwargs,
    ) -> Ax:
        """Plot the mode plane simulation's components.

        Parameters
        ----------
        ax : matplotlib.axes._subplots.Axes = None
            Matplotlib axes to plot on, if not specified, one is created.

        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.

        See Also
        ---------

        **Notebooks**
            * `Visualizing geometries in Tidy3D: Plotting Materials <../../notebooks/VizSimulation.html#Plotting-Materials>`_

        """
        return self._mode_solver.plot(ax=ax, **patch_kwargs)

    def plot_eps_mode_plane(
        self,
        freq: Optional[float] = None,
        alpha: Optional[float] = None,
        ax: Ax = None,
    ) -> Ax:
        """Plot the mode plane simulation's components.
        The permittivity is plotted in grayscale based on its value at the specified frequency.

        Parameters
        ----------
        freq : float = None
            Frequency to evaluate the relative permittivity of all mediums.
            If not specified, evaluates at infinite frequency.
        alpha : float = None
            Opacity of the structures being plotted.
            Defaults to the structure default alpha.
        ax : matplotlib.axes._subplots.Axes = None
            Matplotlib axes to plot on, if not specified, one is created.

        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.

        See Also
        ---------

        **Notebooks**
            * `Visualizing geometries in Tidy3D: Plotting Permittivity <../../notebooks/VizSimulation.html#Plotting-Permittivity>`_
        """
        return self._mode_solver.plot_eps(freq=freq, alpha=alpha, ax=ax)

    def plot_structures_eps_mode_plane(
        self,
        freq: Optional[float] = None,
        alpha: Optional[float] = None,
        cbar: bool = True,
        reverse: bool = False,
        ax: Ax = None,
    ) -> Ax:
        """Plot the mode plane simulation's components.
        The permittivity is plotted in grayscale based on its value at the specified frequency.

        Parameters
        ----------
        freq : float = None
            Frequency to evaluate the relative permittivity of all mediums.
            If not specified, evaluates at infinite frequency.
        alpha : float = None
            Opacity of the structures being plotted.
            Defaults to the structure default alpha.
        cbar : bool = True
            Whether to plot a colorbar for the relative permittivity.
        reverse : bool = False
            If ``False``, the highest permittivity is plotted in black.
            If ``True``, it is plotteed in white (suitable for black backgrounds).
        ax : matplotlib.axes._subplots.Axes = None
            Matplotlib axes to plot on, if not specified, one is created.

        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.

        See Also
        ---------

        **Notebooks**
            * `Visualizing geometries in Tidy3D: Plotting Permittivity <../../notebooks/VizSimulation.html#Plotting-Permittivity>`_
        """
        return self._mode_solver.plot_structures_eps(
            freq=freq, alpha=alpha, cbar=cbar, reverse=reverse, ax=ax
        )

    def plot_grid_mode_plane(
        self,
        ax: Ax = None,
        **kwargs,
    ) -> Ax:
        """Plot the mode plane cell boundaries as lines.

        Parameters
        ----------
        ax : matplotlib.axes._subplots.Axes = None
            Matplotlib axes to plot on, if not specified, one is created.
        **kwargs
            Optional keyword arguments passed to the matplotlib ``LineCollection``.
            For details on accepted values, refer to
            `Matplotlib's documentation <https://tinyurl.com/2p97z4cn>`_.

        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.
        """
        return self._mode_solver.plot_grid(ax=ax)

    def plot_pml_mode_plane(
        self,
        ax: Ax = None,
    ) -> Ax:
        """Plot the mode plane absorbing boundaries.

        Parameters
        ----------
        ax : matplotlib.axes._subplots.Axes = None
            Matplotlib axes to plot on, if not specified, one is created.

        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.
        """
        return self._mode_solver.plot_pml(ax=ax)

    def validate_pre_upload(self, source_required: bool = False):
        self._mode_solver.validate_pre_upload(source_required=source_required)

    _boundaries_for_zero_dims = validate_boundaries_for_zero_dims(warn_on_change=False)



from __future__ import annotations

from collections.abc import Iterable
from functools import reduce, wraps
from typing import Any, Callable, Optional, Union

import autograd.numpy as anp
import numpy as np
import xarray as xr
from numpy.typing import NDArray

from tidy3d.exceptions import Tidy3dError

from .types import KernelType


def _kernel_circular(size: Iterable[int]) -> NDArray:
    """Create a circular kernel in n dimensions.

    Parameters
    ----------
    size : Iterable[int]
        The size of the circular kernel in pixels for each dimension.

    Returns
    -------
    NDArray
        An n-dimensional array representing the circular kernel.
    """
    grids = np.ogrid[tuple(slice(-1, 1, 1j * s) for s in size)]
    squared_distances = sum(grid**2 for grid in grids)
    kernel = np.array(squared_distances <= 1, dtype=np.float64)
    return kernel


def _kernel_conic(size: Iterable[int]) -> NDArray:
    """Create a conic kernel in n dimensions.

    Parameters
    ----------
    size : Iterable[int]
        The size of the conic kernel in pixels for each dimension.

    Returns
    -------
    NDArray
        An n-dimensional array representing the conic kernel.
    """
    grids = np.ogrid[tuple(slice(-1, 1, 1j * s) for s in size)]
    dists = sum(grid**2 for grid in grids)
    kernel = np.maximum(0, 1 - np.sqrt(dists))
    return kernel


def make_kernel(kernel_type: KernelType, size: Iterable[int], normalize: bool = True) -> NDArray:
    """Create a kernel based on the specified type in n dimensions.

    Parameters
    ----------
    kernel_type : KernelType
        The type of kernel to create ('circular' or 'conic').
    size : Iterable[int]
        The size of the kernel in pixels for each dimension.
    normalize : bool = True
        Whether to normalize the kernel so that it sums to 1.

    Returns
    -------
    NDArray
        An n-dimensional array representing the specified type of kernel.
    """
    if not all(np.issubdtype(type(dim), int) and dim > 0 for dim in size):
        raise ValueError("'size' must be an iterable of positive integers.")

    if kernel_type == "circular":
        kernel = _kernel_circular(size)
    elif kernel_type == "conic":
        kernel = _kernel_conic(size)
    else:
        raise ValueError(f"Unsupported kernel type: {kernel_type}")

    if normalize:
        kernel /= np.sum(kernel)

    return kernel


def get_kernel_size_px(
    radius: Optional[Union[float, Iterable[float]]] = None,
    dl: Optional[Union[float, Iterable[float]]] = None,
) -> Union[int, list[int]]:
    """Calculate the kernel size in pixels based on the provided radius and grid spacing.

    Parameters
    ----------
    radius : Union[float, Iterable[float]] = None
        The radius of the kernel. Can be a scalar or an iterable of floats.
    dl : Union[float, Iterable[float]] = None
        The grid spacing. Can be a scalar or an iterable of floats.

    Returns
    -------
    Union[int, List[int]]
        The size of the kernel in pixels for each dimension. Returns an integer if the radius is scalar, otherwise a list of integers.

    Raises
    ------
    ValueError
        If either 'radius' or 'dl' is not provided.
    """
    if radius is None or dl is None:
        raise ValueError("Either 'size_px' or both 'radius' and 'dl' must be provided.")

    if np.isscalar(radius):
        radius = [radius] * len(dl) if isinstance(dl, Iterable) else [radius]
    if np.isscalar(dl):
        dl = [dl] * len(radius)

    radius_px = [np.ceil(r / g) for r, g in zip(radius, dl)]
    return (
        [int(2 * r_px + 1) for r_px in radius_px]
        if len(radius_px) > 1
        else int(2 * radius_px[0] + 1)
    )


def chain(*funcs: Union[Callable, Iterable[Callable]]):
    """Chain multiple functions together to apply them sequentially to an array.

    Parameters
    ----------
    funcs : Union[Callable, Iterable[Callable]]
        A variable number of functions or a single iterable of functions to be chained together.

    Returns
    -------
    Callable
        A function that takes an array and applies the chained functions to it sequentially.

    Examples
    --------
    >>> import numpy as np
    >>> from tidy3d.plugins.autograd.utilities import chain
    >>> def add_one(x):
    ...     return x + 1
    >>> def square(x):
    ...     return x ** 2
    >>> chained_func = chain(add_one, square)
    >>> array = np.array([1, 2, 3])
    >>> chained_func(array)
    array([ 4,  9, 16])

    >>> # Using a list of functions
    >>> funcs = [add_one, square]
    >>> chained_func = chain(funcs)
    >>> chained_func(array)
    array([ 4,  9, 16])
    """
    if len(funcs) == 1 and isinstance(funcs[0], Iterable):
        funcs = funcs[0]

    if not all(callable(f) for f in funcs):
        raise TypeError("All elements in funcs must be callable.")

    def chained(array: NDArray):
        return reduce(lambda x, y: y(x), funcs, array)

    return chained


def scalar_objective(func: Optional[Callable] = None, *, has_aux: bool = False) -> Callable:
    """Decorator to ensure the objective function returns a real scalar value.

    This decorator wraps an objective function to ensure that its return value is a real scalar.
    If the function returns auxiliary data, it expects the return value to be a tuple of the form
    (result, aux_data).

    Parameters
    ----------
    func : Callable, optional
        The objective function to be decorated. If not provided, the decorator should be used with
        arguments.
    has_aux : bool = False
        If True, expects the function to return a tuple (result, aux_data).

    Returns
    -------
    Callable
        The wrapped function that ensures a real scalar return value. If `has_aux` is True, the
        wrapped function returns a tuple (result, aux_data).

    Raises
    ------
    Tidy3dError
        If the return value is not a real scalar, or if `has_aux` is True and the function does not return a tuple of length 2.
    """
    if func is None:
        return lambda f: scalar_objective(f, has_aux=has_aux)

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        result = func(*args, **kwargs)
        aux_data = None

        # Unpack auxiliary data if present
        if has_aux:
            if not isinstance(result, tuple) or len(result) != 2:
                raise Tidy3dError(
                    "If 'has_aux' is True, the objective function must return a tuple of length 2."
                )
            result, aux_data = result

        # Extract data from xarray.DataArray
        if isinstance(result, xr.DataArray):
            result = result.data

        # Squeeze to remove singleton dimensions
        result = anp.squeeze(result)

        # Attempt to extract scalar value
        try:
            result = result.item()
        except AttributeError:
            # If result is already a scalar, pass
            if not isinstance(result, (float, int)):
                raise Tidy3dError(
                    "An objective function's return value must be a scalar, "
                    "a Python float/int, or an array containing a single element."
                ) from None
        except ValueError as e:
            # Result contains more than one element
            raise Tidy3dError(
                "An objective function's return value must be a scalar "
                "but got an array with shape "
                f"{getattr(result, 'shape', 'N/A')}."
            ) from e

        # Ensure the result is real
        if not anp.isreal(result):
            raise Tidy3dError("An objective function's return value must be real.")

        return (result, aux_data) if aux_data is not None else result

    return wrapper



"""Defines sweep settings for the EME simulation."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Union

import pydantic.v1 as pd

from tidy3d.components.base import Tidy3dBaseModel
from tidy3d.components.types import ArrayFloat1D, ArrayInt1D, ArrayLike
from tidy3d.exceptions import SetupError

from .grid import MAX_NUM_REPS


class EMESweepSpec(Tidy3dBaseModel, ABC):
    """Abstract spec for sweep done during EME propagation step."""

    @property
    @abstractmethod
    def num_sweep(self) -> pd.PositiveInt:
        """Number of sweep indices."""


class EMELengthSweep(EMESweepSpec):
    """Spec for sweeping EME cell lengths."""

    scale_factors: ArrayLike = pd.Field(
        ...,
        title="Length Scale Factor",
        description="Length scale factors to be used in the EME propagation step. "
        "The EME propagation step is repeated after scaling every cell length by this amount. "
        "The results are stored in 'sim_data.smatrix'. If a 2D array is provided, the "
        "first index is the sweep index and the second index is the cell index, "
        "allowing a nonuniform cell scaling along the propagation axis.",
    )

    @property
    def num_sweep(self) -> pd.PositiveInt:
        """Number of sweep indices."""
        return len(self.scale_factors)


class EMEModeSweep(EMESweepSpec):
    """Spec for sweeping number of modes in EME propagation step.
    Used for convergence testing."""

    num_modes: ArrayInt1D = pd.Field(
        ...,
        title="Number of Modes",
        description="Max number of modes to use in the EME propagation step. "
        "The EME propagation step is repeated after dropping modes with mode_index "
        "exceeding this value. This can be used for convergence testing; reliable results "
        "should be independent of the number of modes used. This value cannot exceed "
        "the maximum number of modes in any EME cell in the simulation.",
    )

    @property
    def num_sweep(self) -> pd.PositiveInt:
        """Number of sweep indices."""
        return len(self.num_modes)


class EMEFreqSweep(EMESweepSpec):
    """Spec for sweeping frequency in EME propagation step.
    Unlike ``sim.freqs``, the frequency sweep is approximate, using a
    perturbative mode solver relative to the simulation EME modes.
    This can be a faster way to solve at a larger number of frequencies."""

    freq_scale_factors: ArrayFloat1D = pd.Field(
        ...,
        title="Frequency Scale Factors",
        description="Scale factors "
        "applied to every frequency in 'EMESimulation.freqs'. After applying the scale factors, "
        "the new modes are computed approximately using the exact modes as a basis. "
        "If there are multiple 'EMESimulation.freqs', the exact modes are computed at each "
        "of those frequencies, and then the scale factors are applied to each independently.",
    )

    @property
    def num_sweep(self) -> pd.PositiveInt:
        """Number of sweep indices."""
        return len(self.freq_scale_factors)


class EMEPeriodicitySweep(EMESweepSpec):
    """Spec for sweeping number of repetitions of EME subgrids.
    Useful for simulating long periodic structures like Bragg gratings,
    as it allows the EME solver to reuse the modes and cell interface
    scattering matrices.

    Compared to setting ``num_reps`` directly in the ``eme_grid_spec``,
    this sweep spec allows varying the number of repetitions,
    effectively simulating multiple structures in a single EME simulation.

    Example
    -------
    >>> n_list = [1, 50, 100]
    >>> sweep_spec = EMEPeriodicitySweep(num_reps=[{"unit_cell": n} for n in n_list])
    """

    num_reps: list[dict[str, pd.PositiveInt]] = pd.Field(
        ...,
        title="Number of Repetitions",
        description="Number of periodic repetitions of named subgrids in this EME grid. "
        "At each sweep index, contains a dict mapping the name of a subgrid to the "
        "number of repetitions of that subgrid at that sweep index.",
    )

    @pd.validator("num_reps", always=True)
    def _validate_num_reps(cls, val):
        """Check num_reps is not too large."""
        for num_reps_dict in val:
            for value in num_reps_dict.values():
                if value > MAX_NUM_REPS:
                    raise SetupError(
                        f"'EMEGridSpec' has 'num_reps={value:.2e}'; "
                        f"the largest value allowed is '{MAX_NUM_REPS}'."
                    )
        return val

    @property
    def num_sweep(self) -> pd.PositiveInt:
        """Number of sweep indices."""
        return len(self.num_reps)


EMESweepSpecType = Union[EMELengthSweep, EMEModeSweep, EMEFreqSweep, EMEPeriodicitySweep]



from __future__ import annotations

from .axes_utils import add_ax_if_none, equal_aspect, make_ax, set_default_labels_and_title
from .descartes import Polygon, polygon_patch, polygon_path
from .flex_style import apply_tidy3d_params, restore_matplotlib_rcparams
from .plot_params import (
    AbstractPlotParams,
    PathPlotParams,
    PlotParams,
    plot_params_bloch,
    plot_params_fluid,
    plot_params_geometry,
    plot_params_grid,
    plot_params_lumped_element,
    plot_params_monitor,
    plot_params_override_structures,
    plot_params_pec,
    plot_params_pmc,
    plot_params_pml,
    plot_params_source,
    plot_params_structure,
    plot_params_symmetry,
)
from .plot_sim_3d import plot_sim_3d
from .styles import (
    ARROW_ALPHA,
    ARROW_COLOR_MONITOR,
    ARROW_COLOR_POLARIZATION,
    ARROW_COLOR_SOURCE,
    ARROW_LENGTH,
    FLEXCOMPUTE_COLORS,
    MEDIUM_CMAP,
    PLOT_BUFFER,
    STRUCTURE_EPS_CMAP,
    STRUCTURE_EPS_CMAP_R,
    STRUCTURE_HEAT_COND_CMAP,
    arrow_style,
)
from .visualization_spec import MATPLOTLIB_IMPORTED, VisualizationSpec

apply_tidy3d_params()

__all__ = [
    "ARROW_ALPHA",
    "ARROW_COLOR_MONITOR",
    "ARROW_COLOR_POLARIZATION",
    "ARROW_COLOR_SOURCE",
    "ARROW_LENGTH",
    "FLEXCOMPUTE_COLORS",
    "MATPLOTLIB_IMPORTED",
    "MEDIUM_CMAP",
    "PLOT_BUFFER",
    "STRUCTURE_EPS_CMAP",
    "STRUCTURE_EPS_CMAP_R",
    "STRUCTURE_HEAT_COND_CMAP",
    "AbstractPlotParams",
    "PathPlotParams",
    "PlotParams",
    "Polygon",
    "VisualizationSpec",
    "add_ax_if_none",
    "arrow_style",
    "equal_aspect",
    "make_ax",
    "plot_params_bloch",
    "plot_params_fluid",
    "plot_params_geometry",
    "plot_params_grid",
    "plot_params_lumped_element",
    "plot_params_monitor",
    "plot_params_override_structures",
    "plot_params_pec",
    "plot_params_pmc",
    "plot_params_pml",
    "plot_params_source",
    "plot_params_structure",
    "plot_params_symmetry",
    "plot_sim_3d",
    "polygon_patch",
    "polygon_path",
    "restore_matplotlib_rcparams",
    "set_default_labels_and_title",
]



"""Console script subcommand for tidy3d."""

from __future__ import annotations

import click

__all__ = [
    "develop",
]


@click.group(name="develop")
def develop():
    """
    Development related command group in the CLI.

    This command group includes several subcommands for various development tasks such as
    verifying and setting up the development environment, building documentation, testing, and more.
    """
